//
// ACL.cpp
//
// Created by Marcelo Calbucci
//
// April/2000
//
// Compile with _WIN32_WINNT=0x0500 and _UNICODE defined

#include "acl.h"
#include "sd.h"
#include "sderror.h"
#include "winbase.h"
#include <algorithm>

// The operation of reallocate a vector is expensive, so it's worth
// to pre-allocate at least some amount in order to avoid multiple
// allocations later on.
#define ACL_EXTRA_ACES_RESERVE  (16)

/////////////////
// Constructors
/////////////////

// Default constructor
CAcl::CAcl()
{
    m_fModified = FALSE;
    m_pAcl = NULL;
    m_fIsPresent = FALSE;
    m_pParentSD = NULL;
    m_vAces.reserve(ACL_EXTRA_ACES_RESERVE);
}

// Copy constructor
CAcl::CAcl(const CAcl & rhs)
{
    m_fModified = FALSE;
    m_pAcl = NULL;
    m_pParentSD = NULL;
    SetAcl(rhs);
}

// Constructor from a raw ACL
CAcl::CAcl(const PACL pAcl)
{
    m_fModified = FALSE;
    m_pAcl = NULL;
    m_pParentSD = NULL;
    SetAcl(pAcl);
}

// Destructor
CAcl::~CAcl()
{
    delete m_pAcl;
}

///////////////
// Operators
///////////////

// operator assignment
CAcl & CAcl::operator=(const CAcl & rhs)
{
    SetAcl(rhs);
    return *this;
}

// operator equal
BOOL CAcl::operator==(const CAcl & rhs) const
{
    if(&rhs==this) // If we are comparing A==A
        return TRUE;

    if(m_fIsPresent != rhs.m_fIsPresent) // Different presence state
	return FALSE;
    
    if(m_vAces.size() != rhs.m_vAces.size()) // Different number of CAces
	return FALSE;

    BOOL bResult = TRUE;;
    bool * pbRHS = NULL;

    try {
        int iTotalRHS = rhs.m_vAces.size();
        pbRHS = new bool[iTotalRHS];
        ::ZeroMemory(pbRHS, sizeof(bool) * iTotalRHS);
        
        int iPosRHS;
        // Iterate thru each CAce on 'this'
        const_iterator it = begin();
        while(it != end()) {
            iPosRHS = 0;
            while(iPosRHS < iTotalRHS) {
                // Try to find the same CAce on rhs, but it
                // has to be marked as not 'matched' yet.
                if(!pbRHS[iPosRHS]) {
                    if((*it)==rhs.m_vAces[iPosRHS]) {
                        // Found the matching CAce, mark as 'matched'
                        pbRHS[iPosRHS] = TRUE;
                        break;
                    }
                }
                iPosRHS++;
            }
            if(iPosRHS==iTotalRHS) {
                // Couldn't find the matching CAce on rhs.
                bResult = FALSE;
                goto Exit;
            }
            it++;
        }
    } catch(...) {
        delete pbRHS;
        throw;        
    }

Exit:
    delete pbRHS;
    return bResult;
}

// operator not equal
BOOL CAcl::operator!=(const CAcl & rhs) const
{
    return !(operator==(rhs));
}

////////
// Set
////////

// Set from a raw ACL
DWORD CAcl::SetAcl(const PACL pAcl)
{
    if(pAcl==NULL) {
        // Not present
        RemoveAllAces(TRUE);
        return SetSDLastError(SD_SUCCESS);
    } else {
        if(pAcl==m_pAcl) // Self assignment
            return SetSDLastError(SD_SUCCESS);
    }
    RemoveAllAces(TRUE); // Clean up
    
    if(!IsValidAcl(pAcl)) {
	return SetSDLastError(SD_INVALID_ACL);
    }
    
    m_fIsPresent = TRUE;
    if(m_pParentSD)
        m_pParentSD->CheckControlFlags();

    LPVOID pAce;
    DWORD dwError = SD_SUCCESS;

    // Let's reserve the space for the ACEs, plus some extra
    m_vAces.reserve(pAcl->AceCount + ACL_EXTRA_ACES_RESERVE);
    
    // Copy each ACE and add it to the list of CAce.
    int i = 0;
    while(i < pAcl->AceCount) {
	if(::GetAce(pAcl, i, &pAce)) {
            AddAce(pAce);
        } else {
            dwError = SD_GETACE_W32;
        }
	i++;
    }
    return SetSDLastError(dwError);
}

// Set from CAcl
void CAcl::SetAcl(const CAcl & rhs)
{
    if(&rhs==this) // Self assignment
        return;

    RemoveAllAces(TRUE);
    
    if(!rhs.m_fIsPresent)
        return;
    
    // Reserve space
    m_vAces.reserve(rhs.m_vAces.size() + ACL_EXTRA_ACES_RESERVE);

    const_iterator it = rhs.begin();
    while(it != rhs.end()) {
        // Copy each CAce into 'this'.
        // AddAce will also set the 'Parent' CAcl for these CAce
        AddAce(*it);
        it++;
    }
    m_fIsPresent = TRUE;
    if(m_pParentSD)
        m_pParentSD->CheckControlFlags();


}

////////////
// Contain
////////////

// ContainsAcl
//   Verify if the rhs CAcl is represented on this CAcl
// Parameters:
//   rhs: The inner CAcl 
//   bFullMask: TRUE if you want the masks to match
//              FALSE if you want the 'and' of the masks to be equal the rhs mask
//              (default to FALSE)
// Return: TRUE if rhs is represented on this
BOOL CAcl::ContainsAcl(const CAcl & rhs, BOOL bFullMask) const
{
    if(&rhs==this)
        return TRUE;

    if(m_vAces.size() < rhs.m_vAces.size())
	return FALSE;
    
    const_iterator it = rhs.begin();
    while(it != rhs.end()) {
	if(!ContainsAce(*it, bFullMask))
	    return FALSE;
	it++;
    }
    return TRUE;
}

// ContainsAce
//   Verify if the Ace is on this CAcl
// Parameters:
//   Ace: CAce to search for
//   bFullMask: TRUE if you want the mask to be identical (default to FALSE)
// Return: TRUE if the Ace is on this.
BOOL CAcl::ContainsAce(const CAce & Ace, BOOL bFullMask) const
{
    const_iterator it = begin();
    if(bFullMask) {
        // If want full-mask, than the CAce has to be identical
        while(it != end()) {
            if((*it)==Ace)
                return TRUE;
            it++;
        }
    } else {
        // If don't want full-maks, the CAce has to be a sub-set of any other CAce on this
        while(it != end()) {
            if((*it).Contains(Ace))
                return TRUE;
            it++;
        }
    }
    return FALSE;
}

////////
// Add
////////

// AddAce
//   Add an CAce into the list of CAces.
// Parameters:
//   Ace: Ace to be added
//   iPos: Position where to add the new CAce, if less then zero it will be added to the end.
//         (default -1, at the end)
void CAcl::AddAce(const CAce & Ace, int iPos)
{
    iterator itBegin = begin();
    if(iPos < 0 || iPos >= (int)m_vAces.size()) {
	// Insert at the end
	m_vAces.push_back(Ace);
        m_fIsPresent = TRUE;
        if(m_pParentSD)
            m_pParentSD->CheckControlFlags();

        SetModified();
    } else {
	m_vAces.insert(begin() + iPos, Ace);
        // We don't need to call SetModified because at least one CAce
        // we do the notification because it is being moved
    }
    if(itBegin==begin()) {
        // No reallocation was made.
        // The last one is the one that was copied into a new CAce, which means
        // it doesn't have the 'Parent' CAcl set.
        (*(end() - 1)).SetParent(this);
    } else {
        // Every CAce was reallocated to a different memory position
        // Need to set the 'Parent' CAcl;
        itBegin = begin();
        while(itBegin!=end()) {
            (*itBegin).SetParent(this);
            itBegin++;
        }
    }
}

// AddAce
//   Add a raw ACE to the list of CAce
// Parameters:
//   pAce: Pointer to the raw ACE structure
//   iPos: Position to add it (default to -1, at the end)
// Return: Error code
DWORD CAcl::AddAce(const LPVOID pAce, int iPos)
{
    CAce Ace(pAce);
    if(GetSDLastError()==SD_SUCCESS) {
        AddAce(Ace, iPos);
    }
    return GetSDLastError();
}

// AddAccessAllowed
//   Add an Access Allowed ACE with the specified paramater
// Parameters:
//   pSid: Pointer to the SID
//   dwMask: ACE Mask
//   dwAceFlags: ACE Flags (default to 0)
//   guidObjectType: The object-type GUID (default to NULL Guid)
//   guidInheritedObjectType: the inherited-object-type GUID (default to a NULL GUID)
//   iPos: Position to add it (default to -1, at the end)
// Return: Error code
DWORD CAcl::AddAccessAllowed(const PSID pSid, DWORD dwMask, DWORD dwAceFlags, 
			     const GUID & guidObjectType, 
			     const GUID &guidInheritedObjectType,
			     int iPos)
{
    DWORD dw;
    CAce Ace;
    dw = Ace.SetAce(AT_AccessAllowed, pSid, dwMask, dwAceFlags, 
	            guidObjectType, guidInheritedObjectType);
    if(dw==SD_SUCCESS)
	AddAce(Ace, iPos);
    
    return dw;
}

// AddAccessDenied
//   Add an Access Denied ACE with the specified paramater
// Parameters:
//   pSid: Pointer to the SID
//   dwMask: ACE Mask
//   dwAceFlags: ACE Flags (default to 0)
//   guidObjectType: The object-type GUID (default to NULL Guid)
//   guidInheritedObjectType: the inherited-object-type GUID (default to a NULL GUID)
//   iPos: Position to add it (default to -1, at the end)
// Return: Error code
DWORD CAcl::AddAccessDenied(const PSID pSid, DWORD dwMask, DWORD dwAceFlags, 
			    const GUID & guidObjectType, 
			    const GUID &guidInheritedObjectType,
			    int iPos)
{
    DWORD dw;
    CAce Ace;
    dw = Ace.SetAce(AT_AccessDenied, pSid, dwMask, dwAceFlags, 
	guidObjectType, guidInheritedObjectType);
    
    if(dw==SD_SUCCESS)
	AddAce(Ace, iPos);

    return dw;
}

// AddAccessAllowed
//   Add an Access Allowed ACE with the specified paramater
// Parameters:
//   sid: CSid containing the SID
//   dwMask: ACE Mask
//   dwAceFlags: ACE Flags (default to 0)
//   guidObjectType: The object-type GUID (default to NULL Guid)
//   guidInheritedObjectType: the inherited-object-type GUID (default to a NULL GUID)
//   iPos: Position to add it (default to -1, at the end)
void CAcl::AddAccessAllowed(const CSid & sid, DWORD dwMask, DWORD dwAceFlags, 
			    const GUID & guidObjectType, 
			    const GUID &guidInheritedObjectType,
			    int iPos)
{
    CAce Ace;
    Ace.SetAce(AT_AccessAllowed, sid, dwMask, dwAceFlags, 
               guidObjectType, guidInheritedObjectType);
    AddAce(Ace, iPos);
}

// AddAccessDenied
//   Add an Access Denied ACE with the specified paramater
// Parameters:
//   sid: CSid containing the SID
//   dwMask: ACE Mask
//   dwAceFlags: ACE Flags (default to 0)
//   guidObjectType: The object-type GUID (default to NULL Guid)
//   guidInheritedObjectType: the inherited-object-type GUID (default to a NULL GUID)
//   iPos: Position to add it (default to -1, at the end)
void CAcl::AddAccessDenied(const CSid & sid, DWORD dwMask, DWORD dwAceFlags, 
		           const GUID & guidObjectType, 
			   const GUID &guidInheritedObjectType,
			   int iPos)
{
    CAce Ace;
    Ace.SetAce(AT_AccessDenied, sid, dwMask, dwAceFlags, 
	            guidObjectType, guidInheritedObjectType);
    AddAce(Ace, iPos);
}

// AddAuditAccess
//   Add an Audit Access ACE with the specified paramater
// Parameters:
//   pSid: Pointer to the SID
//   dwMask: ACE Mask
//   dwAceFlags: ACE Flags (default to FAILED_ACCESS_ACE_FLAG)
//   guidObjectType: The object-type GUID (default to NULL Guid)
//   guidInheritedObjectType: the inherited-object-type GUID (default to a NULL GUID)
//   iPos: Position to add it (default to -1, at the end)
// Return: Error code
DWORD CAcl::AddAuditAccess(const PSID pSid, DWORD dwMask, DWORD dwAceFlags, 
			   const GUID & guidObjectType, 
			   const GUID &guidInheritedObjectType,
			   int iPos)
{
    if(!(dwAceFlags & (SUCCESSFUL_ACCESS_ACE_FLAG | FAILED_ACCESS_ACE_FLAG))) {
        dwAceFlags = FAILED_ACCESS_ACE_FLAG;
    }

    DWORD dw;
    CAce Ace;
    dw = Ace.SetAce(AT_AuditAccess, pSid, dwMask, dwAceFlags,
	            guidObjectType, guidInheritedObjectType);
    
    if(dw==SD_SUCCESS)
	AddAce(Ace, iPos);
    
    return dw;
}

// AddAuditAccess
//   Add an Audit Access ACE with the specified paramater
// Parameters:
//   sid: CSid containing the SID
//   dwMask: ACE Mask
//   dwAceFlags: ACE Flags (default to FAILED_ACCESS_ACE_FLAG)
//   guidObjectType: The object-type GUID (default to NULL Guid)
//   guidInheritedObjectType: the inherited-object-type GUID (default to a NULL GUID)
//   iPos: Position to add it (default to -1, at the end)
// Return: Error code
void CAcl::AddAuditAccess(const CSid & sid, DWORD dwMask, DWORD dwAceFlags, 
 			  const GUID & guidObjectType, 
 			  const GUID &guidInheritedObjectType,
			  int iPos)
{
    if(!(dwAceFlags & (SUCCESSFUL_ACCESS_ACE_FLAG | FAILED_ACCESS_ACE_FLAG))) {
        dwAceFlags = FAILED_ACCESS_ACE_FLAG;
    }

    CAce Ace;
    Ace.SetAce(AT_AuditAccess, sid, dwMask, dwAceFlags, 
               guidObjectType, guidInheritedObjectType);
    AddAce(Ace, iPos);
}

////////
// Get
////////

// GetAcl
//   Return a reference pointer to an ACL
// Return: Pointer to the ACL structure
const PACL CAcl::GetAcl()
{
    return GetAcl(0);
}

// GetAcl (protected)
//   Return a reference pointer to an  ACL
// Parameters:
//   iExtraSpace: Extra space to be allocated on the ACL (default to 0)
// Return: Pointer to the ACL structure.
const PACL CAcl::GetAcl(DWORD iExtraSpace)
{
    if(m_pAcl) {
        if(m_fModified) {
            // The CAcl has been modified since the last time we generated m_pAcl
            delete m_pAcl;
            m_pAcl = NULL;
        } else {
            if(m_pAcl->AclSize < (iExtraSpace + LengthRequired())) {
                // We do have the pointer but its size is less than what the caller ask for
                delete m_pAcl;
                m_pAcl = NULL;
            } else {
                // If we have the pointer and it is block of memory is the right size
                // return it.
                return m_pAcl;
            }
        }
    }

    SetSDLastError(SD_SUCCESS);

    if(!m_fIsPresent) {
	return NULL; // Not present means NULL ACL
    }
    
    DWORD dwSize = LengthRequired() + iExtraSpace;
    m_pAcl = (PACL)new UCHAR[dwSize];
    if(!m_pAcl) {
	SetSDLastError(SD_OUT_OF_MEMORY);
	return NULL;
    }
    try {
        // Initialize the ACL structure
        if(!InitializeAcl(m_pAcl, dwSize, ACL_REVISION_DS)) {
            delete m_pAcl;
            m_pAcl = NULL;
            SetSDLastError(SD_INITIALIZE_ACL_W32);
            return NULL;
        }
        
        // Copy the ACEs into the ACL
        DWORD dwError = SD_SUCCESS;
        const_iterator it = begin();
        while(it != end()) {
            dwError = (*it).AddIntoAcl(m_pAcl);
            if(dwError!=SD_SUCCESS) {
                delete m_pAcl;
                m_pAcl = NULL;
                return NULL;
            }
            it++;
        }
    } catch(...) {
        delete m_pAcl;
        m_pAcl = NULL;
        throw;
    }
    m_fModified = FALSE;

    return m_pAcl;
}

// GetAclCopy
//   Get a copy of the ACL structure. You need to call DeleteAcl to delete this pointer
// Parameter:
//   iExtraSpace: Extra space in bytes to be added to the memory allocation
// Return: The pointer to the ACL
PACL CAcl::GetAclCopy(DWORD iExtraSpace)
{

    PACL pAcl = GetAcl(iExtraSpace);
    if(pAcl==NULL)
	return NULL;
    
    PACL pAclCopy = (PACL)new UCHAR[pAcl->AclSize];
    if(!pAclCopy) {
	SetSDLastError(SD_OUT_OF_MEMORY);
	return NULL;
    } 
    memcpy(pAclCopy, pAcl, pAcl->AclSize);
    SetSDLastError(SD_SUCCESS);
    return pAclCopy;
}

// Ace
//   Return a reference to the ACE at position iPos. 
//   It is safe to edit the content of CAce.
// Parameter:
//   iPos: Position of the ACE to be retrivied. If iPos is invalid, an empty ACE is returned.
CAce & CAcl::Ace(int iPos)
{
    static CAce EmptyAce;
    if(iPos < (int)m_vAces.size()) {
	return (*(begin() + iPos));
    } else {
        EmptyAce.Clear(); // We need to clear it up because somebody might have edit it.
	return EmptyAce;
    }
}


///////////////////
// Delete/Remove
///////////////////

// RemoveAce
//   Remove an ACE from the list and get a copy of the content.
// Parameters:
//   iPos: Position where you want the CAce to be removed.
//   Ace (out): Receive the CAce being removed
// Return: TRUE if the position was valid and CAce was removed.
BOOL CAcl::RemoveAce(int iPos, CAce & Ace)
{
    Ace.Clear();
    if(iPos >= 0 && iPos < (int)m_vAces.size()) {
        Ace = *(begin() + iPos);
        m_vAces.erase(begin() + iPos);
        SetModified();
        return TRUE;
    }
    return FALSE;
}

// DeleteAce
//   Delete an ACE from the list
// Parameter:
//   iPos: Poisiton where to remove the CAce.
void CAcl::DeleteAce(int iPos)
{
    if(iPos < (int)m_vAces.size()) {
	m_vAces.erase(begin() + iPos);
        SetModified();
    }
}

// DeleteAce
//   Delete an ACE from the list
// Parameter:
//   it: iterator to the CAce position.
void CAcl::DeleteAce(iterator it)
{
    if(it >= begin() && it < end()) {
        m_vAces.erase(it);
        SetModified();
    }
}

// DeleteAcl (static)
//   Delete the content allocated by GetAclCopy().
// Parameter:
//   pAcl: Pointer to the ACL returned by GetAclCopy().
void CAcl::DeleteAcl(PACL pAcl)
{
    delete pAcl;
}

//////////////
// Specials
//////////////

// RemoveAllAces
//   Clean up the entire CAcl structure.
// Parameter:
//   bMakeItNotPresent: If TRUE, the ACL is considered not present.
void CAcl::RemoveAllAces(BOOL bMakeItNotPresent)
{
    m_fIsPresent = !bMakeItNotPresent;
    m_vAces.clear();
    if(m_pParentSD)
        m_pParentSD->CheckControlFlags();
    SetModified();
}

// CanonicalSort
//   Sort the ACEs in a Canonical format.
void CAcl::CanonicalSort()
{
    SetSDLastError(SD_SUCCESS);
    std::sort(begin(), end(), CCanonicalSort());
    SetModified();
}

// RemoveInheritedAces()
//   Remove all ACEs that have the INHERITED_ACE bit set.
void CAcl::RemoveInheritedAces()
{
    SetSDLastError(SD_SUCCESS);
    BOOL fModified = FALSE;
    iterator it = begin();
    while(it != end()) {
        if((*it).IsInherited()) {
	    m_vAces.erase(it);
            fModified = TRUE;
        } else {
	    it++;
        }
    }
    if(fModified)
        SetModified();
}

//////////
// Misc
//////////

// AceCount
//   Return the number of CAce on this CAcl
DWORD CAcl::AceCount() const
{
    return m_vAces.size();
}

// LengthRequired
//   Return the length that will be used by the ACL structure to represent this CAcl (in bytes)
DWORD CAcl::LengthRequired() const
{
    DWORD dwTotal = sizeof(ACL);
    const_iterator it = begin();
    while(it != end()) {
	dwTotal += (*it).LengthRequired();
	it++;
    }
    return dwTotal;
}

// SetModified (protected)
//   Set the modified bit and propagate it to the parent CSD.
void CAcl::SetModified()
{
    if(m_fModified)
        return;

    m_fModified = TRUE;
    if(m_pParentSD)
        m_pParentSD->SetModified();
}

///////////////////
// Canonical Sort
///////////////////

int CCanonicalSort::operator()(const CAce & rhs, const CAce & lhs)
{
    return GetOrder(lhs) - GetOrder(rhs);
};

int CCanonicalSort::GetOrder(const CAce & rhs)
{
    int iOrder = 0;
    if(!rhs.IsInherited())
        iOrder += 0x1000;
    if(rhs.IsDenied())
        iOrder += 0x100;
    if(!rhs.IsObjectTypeAce())
        iOrder += 0x10;
    return iOrder;
}
