// PropFindCmd.cpp: implementation of the CPropFindCmd class.
//
//////////////////////////////////////////////////////////////////////

/*
 *	Martin Hallerdal, marty@home.se
 *	2000-05-11
 *
 *	Please feel free to experiment with the code. Just don't use it for commercial purposes.
 */


#include "stdafx.h"

#include "PropFindCmd.h"
#include "DavCmd.h"
#include "XMLBody.h"
#include "Props.h"
#include "HttpUtil.h"

#include <limits.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


//PropSpec implementation

void PropSpec::AddProperty(char *szName) {
	
	//Check if it is a DAV-property

	bool bIsDav = false;

	for(int i = 0; i < dwNumProps; i++) {
		if(strcmp(szName, davProps[i].szName) == 0) {

			dProps.push_back(davProps[i]);
			bIsDav = true;
		}
	}
		
	//If not a standard DAV property, add it to the list
	
	if(!bIsDav) {
		int iLen = strlen(szName);
		
		char *szNy = new char[iLen+1];
		strcpy(szNy, szName);
		szNy[iLen] = '\0';
		
		props.push_back(szNy);
	}
}

PropSpec::~PropSpec() {
	std::vector<char*>::iterator it;
	
	for(it = props.begin(); it != props.end(); it++) {
		delete *it;
	}
}



const DavProperty PropSpec::davProps[] = {
	{"creationdate", DAV_CREATIONDATE_PROP},
	{"displayname", DAV_DISPLAYNAME_PROP},
	{"getcontentlanguage", DAV_GETCONTENTLANGUAGE_PROP},
	{"getcontentlength", DAV_GETCONTENTLENGTH_PROP},
	{"getcontenttype", DAV_GETCONTENTTYPE_PROP},
	{"getetag", DAV_GETETAG_PROP},
	{"getlastmodified", DAV_GETLASTMODIFIED_PROP},
	{"lockdiscovery", DAV_LOCKDISCOVERY_PROP},
	{"resourcetype", DAV_RESOURCETYPE_PROP},
	{"source", DAV_SOURCE_PROP},
	{"supportedlock", DAV_SUPPORTEDLOCK_PROP}
};

const DWORD PropSpec::dwNumProps = 11;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPropFindCmd::CPropFindCmd(CDavCmdDispatcher* disp, CHttpFilterContext* pCtx) : CDavCmd(disp, pCtx) {

}

CPropFindCmd::~CPropFindCmd() {

}

/*
	Gets the base URL 
*/

void CPropFindCmd::GetBaseURL() {
	
	int iProtoLen;

	if(pCtx->m_pFC->fIsSecurePort) {
		strcpy(szURL,"https://");
		iProtoLen = 8;
	} else {
		strcpy(szURL,"http://");
		iProtoLen = 7;
	}
	
	DWORD dwAvail = 512 - iProtoLen;
	pCtx->GetServerVariable("SERVER_NAME", szURL + iProtoLen, &dwAvail);

	iServerLen = iProtoLen + dwAvail - 1;

}

/*
	Executes the PROPFIND command
*/

bool CPropFindCmd::Execute() {
	
	//Get the server url

	GetBaseURL();

	//Get the request body
	
	IXMLDOMElementPtr elemRoot = GetRootXMLElement();
	
	//If elemRoot is NULL it means that the request body was empty.
	//Assume that we want names and values of all properties

	if(elemRoot != NULL) {

		//Check if this contains the "propfind"  element

		if(strstr(elemRoot->baseName, "propfind") == NULL) {
			//**TODO . Send an error message

			return false;
		}

		//Find out what properties we want

		GetPropSpecs(elemRoot);
	} else {

		//No request body. Assume all properties

		for(int i = 0; i < PropSpec::dwNumProps; i++) {
			specs.dProps.push_back(PropSpec::davProps[i]);
		}
	}

	//Determine depth

	const char* szDepth = pDisp->pHdrs->GetHeader("Depth");

	if(szDepth != NULL) {
		
		int iDepth = atoi(szDepth);

		//If we can convert to a number
		
		if(iDepth == 0) {
			specs.depth = ZERO;
		} else if(iDepth == 1) {
			specs.depth = ONE;
		}

	}
	

	//Ok. Build response from the specifications

	BuildResponse();

	//Send it to the client

	resp.SendResponse(pCtx);

	return true;
}


/*
	Creates a PROPFIND response based on the specifications
*/

void CPropFindCmd::BuildResponse() {

	//IXMLDOMElementPtr nextElemRoot;
	
	//Create the document

	IXMLDOMDocumentPtr doc;
	doc.CreateInstance("MSXML.DOMDocument");

	//Create a root element
	
	IXMLDOMElementPtr elemRoot = doc->createNode(1L, "D:multistatus", "DAV:");

	//Build the tree
	
	this->elemRoot = elemRoot;
	this->doc = doc;

	BuildResponseTree(true, pDisp->szPhys, pDisp->szURI);

	
	//Only if we have a body should the response be 207 Multistatus

	if(bDidInit) {

		//Set the element as the root
		
		doc->documentElement = elemRoot;
		
		//Create the body

		CXMLBody* pResp = new CXMLBody;
		pResp->SetXMLDocument(doc);
		
		resp.SetStatus(207);
		resp.pBody = pResp;
		
		_bstr_t bstrXml = doc->xml;
		const char* szTemp = (char*)bstrXml;
	}



	

	
}


/*
	Creates a response element in elemRoot for the resource szURI

	Typical response

	<response>
		<href></href>
		
		 <propstat>
			<prop></prop>
			<status></status>
		</propstat>
		
		<propstat>
		....
	</response>
	
			
*/

void CPropFindCmd::BuildResponseTree(bool bFirst, char* szPhysLoc, char* szURI) {

	DWORD dwPhysLen = strlen(szPhysLoc);
	DWORD dwUriLen = strlen(szURI);

	//Fix szPhysLoc and szURI if they have a trailing \ or /
	
	if(szPhysLoc[dwPhysLen-1] == '\\') {
		szPhysLoc[dwPhysLen-1] = '\0';
		dwPhysLen--;
	}

	if(szURI[dwUriLen-1] != '/') {
		szURI[dwUriLen] = '/';
		szURI[dwUriLen+1] = '\0';
		dwUriLen++;
	}

	//Add the URI to the URL

	szURL[iServerLen] = '\0';
	strcat(szURL, szURI);

	//Create a response element

	IXMLDOMNodePtr elemResponse = doc->createElement("D:response");

	//Create a href element and add it

	IXMLDOMNodePtr elemHref = doc->createElement("D:href");

	IXMLDOMTextPtr elemText = doc->createTextNode(szURI);
	elemHref->appendChild(elemText);

	elemResponse->appendChild(elemHref);	


	//Get some file attributes
	
	GET_FILEEX_INFO_LEVELS lev = GetFileExInfoStandard;
	WIN32_FILE_ATTRIBUTE_DATA attrs;

	DWORD dwAttrs = ::GetFileAttributesEx(szPhysLoc, lev, &attrs);
	
	if(dwAttrs == 0) {
		//Something went wrong. Create a propstat for every property
		
		if(bFirst) {
			SetError(true, szURI);
		} else {
			bDidInit = true;
			GetErrorPropStat(elemResponse);
		}
		
			
	} else {

		bDidInit = true;
	
		//Success, find properties

		//Get propstat for all unsupported properties

		GetUnsupportedPropStat(elemResponse);

		//Get a propstat for the DAV properties requested

		GetDavPropStat(elemResponse, szPhysLoc, &attrs, dwPhysLen);

		//If this is a directory, change the URL

		if(! (attrs.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
			_bstr_t newURL = elemText->substringData(0, elemText->length - 1);
			elemText->deleteData(0, elemText->length);
			elemText->appendData(newURL);
		}

		if(attrs.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY && specs.depth != ZERO) {

			//Prepare specs for collections by decreasing depth
			
			if(specs.depth = ONE) {
				specs.depth = ZERO;
			}
			
			//List files
			
			char szSearchSpec[MAX_PATH];
			strcpy(szSearchSpec, szPhysLoc);
			strcat(szSearchSpec, "\\*.*");

			WIN32_FIND_DATA fdata;
			BOOL bFind;			
			
			HANDLE hFind = ::FindFirstFile(szSearchSpec, &fdata);
			
			if(hFind != INVALID_HANDLE_VALUE) {

				bDidInit = true;

				while(true) {
					
					//Do stuff
				
					if(strcmp(fdata.cFileName, ".") && strcmp(fdata.cFileName, "..") && !(fdata.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN)) {
	

						szPhysLoc[dwPhysLen] = '\\';
						szPhysLoc[dwPhysLen+1] = '\0';

						szURI[dwUriLen] = '\0';

						strcat(szPhysLoc, fdata.cFileName);
						strcat(szURI, fdata.cFileName);

						BuildResponseTree(false, szPhysLoc, szURI);
					}

					
				
					//Get the next file
					
					bFind = ::FindNextFile(hFind, &fdata);
					
					if(!bFind) {
						if(GetLastError() == ERROR_NO_MORE_FILES) {
							::FindClose(hFind);	
							break;
						}
					} 
				
				} //Loop files

				
			} else {
				//Invalid handle
				bDidInit = false;
				SetError(bFirst, szURI);
			}

		
		} //Depth = (ONE or INFINITY)



		
	} //Successfully got properties

	//Add the response

	if(bDidInit) {
		elemRoot->appendChild(elemResponse);
	}
}

/*
	Some operation has failed. Find out what it is and create a 
	propstat for all the properties requested
*/
void CPropFindCmd::GetErrorPropStat(IXMLDOMNodePtr elemResponse) {

	IXMLDOMNodePtr propStat;
	IXMLDOMNodePtr prop;
	IXMLDOMNodePtr status;
	IXMLDOMTextPtr statusText;
	char* szStatus = ::GetHTTPStatusLine();
	
	//Create <propstat>
	
	propStat = doc->createElement("D:propstat");
	
	//Create <prop>

	prop = doc->createElement("D:prop");
	
	//Create <status>
	
	status = doc->createElement("D:status");
	
	//Set the appropriate status
	
	statusText = doc->createTextNode(szStatus);
	status->appendChild(statusText);
	
	//Set the properties for which this status applies. All actually.
	
	IXMLDOMElementPtr elemProp;
	
	std::vector<DavProperty>::iterator iDav;
	
	for(iDav = specs.dProps.begin(); iDav != specs.dProps.end(); iDav++) {
		elemProp = doc->createElement((*iDav).szName);			
		prop->appendChild(elemProp);
	}
	
	std::vector<char*>::iterator iOth;
	
	for(iOth = specs.props.begin(); iOth != specs.props.end(); iOth++) {
		elemProp = doc->createElement((*iOth));			
		prop->appendChild(elemProp);
	}
	
	
	propStat->appendChild(prop);
	propStat->appendChild(status);
	elemResponse->appendChild(propStat);
}

/*
	Adds all the properties that we don't handle to a propstat in the response
*/

void CPropFindCmd::GetUnsupportedPropStat(IXMLDOMNodePtr elemResponse) {
	if(specs.props.begin() != specs.props.end()) {
		
		
		IXMLDOMNodePtr unSupPropStat = doc->createElement("D:propstat");
		IXMLDOMNodePtr unSupProp = doc->createElement("D:prop");
		IXMLDOMNodePtr unSupStatus = doc->createElement("D:status");
		IXMLDOMNodePtr prop;
		IXMLDOMTextPtr statusText;
		
		std::vector<char*>::iterator iOth;
		
		for(iOth = specs.props.begin(); iOth != specs.props.end(); iOth++) {
			prop = doc->createElement((*iOth));			
			unSupProp->appendChild(prop);
		}
		
		statusText = doc->createTextNode("HTTP/1.1 404 Not Found");
		unSupStatus->appendChild(statusText);
		
		unSupPropStat->appendChild(unSupProp);
		unSupPropStat->appendChild(unSupStatus);
		elemResponse->appendChild(unSupPropStat);
		
	}
	
}


/*
	Adds DAV properties to a specified response element for a certain resource
*/

BOOL CPropFindCmd::GetDavPropStat(IXMLDOMNodePtr elemResponse , char* szPhys, WIN32_FILE_ATTRIBUTE_DATA* attrs, DWORD& dwPhysLen) {
	
	IXMLDOMTextPtr elemPropText;		//Text value of a property, e.g. 1999-12-12
	IXMLDOMNodePtr propName;			//Name of a property, e.g. creationdate

	IXMLDOMTextPtr statusText;
	IXMLDOMNodePtr elemPropStat = doc->createElement("D:propstat");
	IXMLDOMNodePtr elemProp = doc->createElement("D:prop");
	IXMLDOMNodePtr elemStatus = doc->createElement("D:status");
	

	statusText = doc->createTextNode("HTTP/1.1 200 OK");
	elemStatus->appendChild(statusText);


	//Get the display name

	char* szDispName = szPhys + dwPhysLen - 2;
	
	while(*szDispName != '\\') {
		szDispName--;
	}
			
	szDispName++;
			
	

	//Iterate through the properties

	if(!specs.bJustNames) {
		
		char szValBuff[512];
		BOOL bRetVal;
		
		
		std::vector<DavProperty>::iterator iDav;
		char szElemName[128] = "D:";
		
		for(iDav = specs.dProps.begin(); iDav != specs.dProps.end(); iDav++) {
			
			strcpy(szElemName+2,(*iDav).szName); 		
			propName = doc->createElement(szElemName);
			
			switch((*iDav).dwVal) {
				
			case DAV_CREATIONDATE_PROP:			//BUG In GetISO8601Time!
				
				//Get the creation date property
				
				bRetVal = ::GetISO8601Time(attrs->ftCreationTime, szValBuff);
				
				if(bRetVal) {
					elemPropText = doc->createTextNode(szValBuff);
					propName->appendChild(elemPropText);
				}
				
				
				break;
				
			case DAV_DISPLAYNAME_PROP:			//DONE!
				
				//Basically the name of the directory or file
				
				elemPropText = doc->createTextNode(szDispName);
				propName->appendChild(elemPropText);
				
				
				break;
			case DAV_GETCONTENTLANGUAGE_PROP:	//IGNORE - VALID
				
				
				
				break;
				
			case DAV_GETCONTENTLENGTH_PROP:		//DONE!
				
				//Length. Note, this limits the file length to 4GB
				
				sprintf(szValBuff, "%d", attrs->nFileSizeLow);
				elemPropText = doc->createTextNode(szValBuff);
				propName->appendChild(elemPropText);
				
				break;
				
			case DAV_GETCONTENTTYPE_PROP:		//DONE!
				
				
				::GetMimeType(szDispName, szValBuff);
				
				elemPropText = doc->createTextNode(szValBuff);
				propName->appendChild(elemPropText);
				
				
				break;
				
			case DAV_GETETAG_PROP:				//IGNORE - IS IT VALID? NO!
				
				break;
				
			case DAV_GETLASTMODIFIED_PROP:		//DONE!
				
				bRetVal = ::GetISO8601Time(attrs->ftLastWriteTime, szValBuff);
				
				if(bRetVal) {
					elemPropText = doc->createTextNode(szValBuff);
					propName->appendChild(elemPropText);
				}
				
				break;
				
			case DAV_LOCKDISCOVERY_PROP:		//IGNORE - VALID
				
				//If we support locking, check for locks here
				
				break;
				
			case DAV_RESOURCETYPE_PROP:			//DONE!
				
				//Directory? Make a collection, otherwise don't do anything
				
				if(attrs->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
					propName->appendChild(doc->createElement("D:collection"));
				}
				
				break;
				
			case DAV_SOURCE_PROP:				//IGNORE - VALID
				break;
				
			case DAV_SUPPORTEDLOCK_PROP:		//IGNORE - VALID
				break;
			}
			
			elemProp->appendChild(propName);
		}
		
	} else {
		//Just add names

		for(int i = 0; i < PropSpec::dwNumProps; i++) {

			elemPropText = doc->createTextNode(PropSpec::davProps[i].szName);
			elemPropStat->appendChild(elemPropText);
			
		}
		
	}

	elemPropStat->appendChild(elemProp);
	elemPropStat->appendChild(elemStatus);

	elemResponse->appendChild(elemPropStat);

	return TRUE;
}


/*
	Retrieves specifications about what properties to collect and how deep we should go.
	The result winds up in specs
*/

bool CPropFindCmd::GetPropSpecs(IXMLDOMElementPtr elemReq) {
	
	IXMLDOMElementPtr elemCurr = elemReq->firstChild;
	
	if(elemCurr == NULL) {
		specs.bAllProps = true;
		return false;
	}
	
	_bstr_t bstrElemName = elemCurr->baseName;
	char* szElemName = (char*)bstrElemName;
	
	//All properties
	
	if(strstr(szElemName, "allprop") != NULL) {

		//Add all the properties we know of

		for(int i = 0; i < PropSpec::dwNumProps; i++) {
			specs.dProps.push_back(PropSpec::davProps[i]);
		}
		
		specs.bAllProps = true;
		return true;
	}

	//All names

	if(strstr(szElemName, "propname") != NULL) {
		specs.bJustNames = true;
		return true;
	}

	//Specific properties

	if(strstr(szElemName, "prop") != NULL) {

		//Get a list of child nodes, one node for each property

		IXMLDOMNodeListPtr elemList = elemCurr->childNodes;
		IXMLDOMNodePtr nodeCurr;

		if(elemList == NULL) {
			specs.bAllProps = true;
		}

		specs.bAllProps = false;
		specs.bJustNames = false;

		long lLength = elemList->length;
	
		for(int i=0; i  < lLength; i++) {
			nodeCurr = elemList->item[i];

			if(nodeCurr != NULL) {
				bstrElemName = nodeCurr->baseName;
				szElemName = (char*)bstrElemName;

				//Add the property to the list

				specs.AddProperty(szElemName);
			}

		}

	}

	return true;
}
