// RequestInfo.cpp: implementation of the CRequestInfo class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "DavFilter.h"
#include "RequestInfo.h"
#include "HttpConst.h"
#include "Headers.h"
#include "DavCmdDispatcher.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*
 *	Martin Hallerdal, marty@home.se
 *	2000-05-11
 *
 *	Please feel free to experiment with the code. Just don't use it for commercial purposes.
 */


CRequestInfo::CRequestInfo() : dwInputLen(0), pcInput(NULL), bForceReadMode(false), szPhysLocation(NULL), pcLastBlock(NULL), bShouldIgnoreRequest(false), bOptionsRequest(false)
{

}

CRequestInfo::~CRequestInfo()
{
	SAFE_FREE_ARRAY(pcInput);
	SAFE_FREE_ARRAY(pcLastBlock);

}

/*
	Adds data to this request
*/

void CRequestInfo::AddData(PHTTP_FILTER_RAW_DATA theData)
{
	//If we are processing a DAV command we should save data
	
	if(!bShouldIgnoreRequest) {
		StackLastBlock();
	}

	//For every type of request we must save the last block of data.

	SetLastBlock((char*)theData->pvInData, theData->cbInData);
}

/*
	Returns the amount of data we have got
*/

DWORD CRequestInfo::GetBufferSize()
{
	return dwInputLen + dwLastLen;
}

/*
	Indicates that the last block we received was the headers.
	Returns the amount of body data in the header block
*/

DWORD CRequestInfo::ReceivedHeaders(char *szMethod, char* szURI)
{

	long lRetval;

	//The last block of data we received was the headers. Headers always start a new request
	
	SAFE_FREE_ARRAY(pcInput);
	dwInputLen = 0;
	
	StackLastBlock();

	//Parse headers

	CHeaders* hdrs = CHeaders::ParseHeaders(pcInput, dwInputLen, &lRetval);

	davCmd.SetMethod(szMethod);
	davCmd.SetHeaders(hdrs);
	davCmd.SetURI(szURI);

	bShouldIgnoreRequest = davCmd.IsDavCommand() ? false : true;
	
	return lRetval;
}



/*
	If we have a web-dav command, process it
*/

void CRequestInfo::ProcessCommand(CHttpFilterContext* pCtx)
{

	//Add the last block of data

	StackLastBlock();

	//Find the body border

	DWORD dwBodyLen;
	char* szHeaderDelim = strstr(pcInput, "\r\n\r\n");

	if(szHeaderDelim == NULL) {
		dwBodyLen = 0;

	} else {

		szHeaderDelim += 4;
		dwBodyLen = dwInputLen - (szHeaderDelim-pcInput);
	}

	//Set the command body

	davCmd.SetCommandBody(szHeaderDelim, dwBodyLen);
	davCmd.Execute(pCtx);
	
	Restore();
}

/* 
	Sets the physical location for this resource.
*/

void CRequestInfo::SetPhysicalLocation(const char *plc, int len)
{
	davCmd.SetPhysicalLocation(plc,len);
}

/*
	Adds the content of the pcLastBlock to the end of pcInput 
*/

void CRequestInfo::StackLastBlock()
{
	if(dwLastLen != 0 && pcLastBlock != NULL) {

		DWORD dwNewLen = dwInputLen + dwLastLen + 1;
		
		char* nyData = new char[dwNewLen];
		
		//Copy the existing data
		
		memcpy(nyData, pcInput, dwInputLen);
		
		nyData += dwInputLen;
		memcpy(nyData, pcLastBlock, dwLastLen);
		
		nyData -= dwInputLen;
		
		SAFE_FREE_ARRAY(pcInput);
		
		nyData[dwNewLen-1] = '\0';
	
		pcInput = nyData;
		
		//Increment size pointer
		
		dwInputLen += dwLastLen;
		
		SAFE_FREE_ARRAY(pcLastBlock);

		dwLastLen = 0;
	}
}

/*
	Sets the content of the last block - pcLastBlock 
*/

void CRequestInfo::SetLastBlock(char *strData, int len)
{
	//No buffer at all?
	
	if(pcLastBlock == NULL) {
		pcLastBlock = new char[len];
		dwLastLen = len;
	}

	//Do we have to allocate a bigger buffer?
	
	if(dwLastLen < len) {
		SAFE_FREE_ARRAY(pcLastBlock);
		pcLastBlock = new char[len];
		dwLastLen = len;
	}
	
	dwLastLen = len;
	memcpy(pcLastBlock, strData, dwLastLen);
}

/*
	Resets this request to an uninitialized state
*/

void CRequestInfo::Restore()
{

	SAFE_FREE_ARRAY(pcLastBlock);
	SAFE_FREE_ARRAY(pcInput);

	lBytesToGo = 0;
	dwInputLen = 0;
	dwLastLen = 0;
	bForceReadMode = false;
	bShouldIgnoreRequest = false;
	bOptionsRequest = false;
}

/*
	Returns true if this is a DAV command we're handling
*/

bool CRequestInfo::IsDavCommand() {
	return davCmd.IsDavCommand();
}


/*
	Stand-in function for the filter

	This method is called after a number of raw reads and after the server has pre-processed
	the headers. We tell our request object that the last block he received was the headers.
	We then ask him if that was a DAV command and if not, we quit.
	
	We then look for a Content-Length header which indicates that more data apart
	from the headers is coming. 

*/

DWORD CRequestInfo::OnPreprocHeaders(CHttpFilterContext *pCtxt, PHTTP_FILTER_PREPROC_HEADERS pHeaderInfo) {
	
	//Get the method in question
	
	char szMethod[20];
	DWORD dwMethLen = 20;

	BOOL bRetVal = pHeaderInfo->GetHeader(pCtxt->m_pFC, "method", szMethod, &dwMethLen);
	
	if(bRetVal == FALSE) {
		strcpy(szMethod, "");
	}

	//Get the URI

	char szURI[512];
	DWORD dwURILen = 20;

	bRetVal = pHeaderInfo->GetHeader(pCtxt->m_pFC, "url", szURI, &dwURILen);
	
	if(bRetVal == FALSE) {
		strcpy(szURI, "");
	}

	//Check if options command


	if(!strcmp("OPTIONS", szMethod)) {
		bOptionsRequest = true;

	}
	
	//We've received the headers

	DWORD dwBodyBytes = ReceivedHeaders(szMethod, szURI);
	
	//If it isn't a DAV command, don't bother continuing

	if(!IsDavCommand()) {
		return SF_STATUS_REQ_NEXT_NOTIFICATION;
	}

	//Get content-length header. If there exists a header Content-Length there is more data.

	char szHdr[20];
	unsigned long lSize = 20;
	
	bRetVal = pHeaderInfo->GetHeader(pCtxt->m_pFC, "Content-Length:", szHdr, &lSize);

	if(bRetVal == TRUE) {

		//There exists a content-length header. More data to come.

		szHdr[lSize] = '\0';
		unsigned long lConLen = atoi(szHdr);
		
		//Subtract the number of body bytes we already have

		lConLen -= dwBodyBytes;

		lBytesToGo = lConLen;

	} else {
		lBytesToGo = 0;
	}

	return SF_STATUS_REQ_NEXT_NOTIFICATION;
}

/*
	Stand-in function for the filter
*/

DWORD CRequestInfo::OnUrlMap(CHttpFilterContext *pCtxt, PHTTP_FILTER_URL_MAP pMapInfo) {

	//Quit if we have been here before or should ignore command
	
	if(bForceReadMode || bShouldIgnoreRequest) {
		return SF_STATUS_REQ_NEXT_NOTIFICATION;
	}

	SetPhysicalLocation(pMapInfo->pszPhysicalPath, pMapInfo->cbPathBuff);

	//If we have more data to read, make one final call

	if(lBytesToGo != 0) {
		unsigned long lToGo = lBytesToGo;

		BOOL bRetval = pCtxt->ServerSupportFunction(SF_REQ_SET_NEXT_READ_SIZE, NULL, &lToGo, &lToGo);
		
		if(bRetval == FALSE) {
			//What to do here?
		}

		//Put the request in forced read mode and tell IIS to make another call
		
		bForceReadMode = true;	

		return SF_STATUS_REQ_READ_NEXT;

	} else {

		//No more data is coming. Process command

		ProcessCommand(pCtxt);

		return SF_STATUS_REQ_FINISHED_KEEP_CONN;
		//return SF_STATUS_REQ_FINISHED;
	}
	
	return SF_STATUS_REQ_NEXT_NOTIFICATION;
}



/*
	Stand-in function for the filter
*/

DWORD CRequestInfo::OnReadRawData(CHttpFilterContext *pCtxt, PHTTP_FILTER_RAW_DATA pRawData)
{

	//Add data to the request

	AddData(pRawData);

	if(bForceReadMode) {
		//Assume that the number of bytes we requested in OnUrlMap has been read

		ProcessCommand(pCtxt);
		
		return SF_STATUS_REQ_FINISHED_KEEP_CONN;
		//return SF_STATUS_REQ_FINISHED;
	}

	return SF_STATUS_REQ_NEXT_NOTIFICATION;
}


/*
	Stand-in function for the filter

	BUG!! "OPTIONS *" causes the function to fuck up
*/

DWORD CRequestInfo::OnSendRawData(CHttpFilterContext* pCtxt,	
								PHTTP_FILTER_RAW_DATA pRawData) {

	//Modify headers if this is an OPTIONS response
	
	static const char* szNewHeader = "DAV: 1\r\n"; 
	static const char* szAnotherNewHeader = "MS-Author-Via: DAV\r\n";
	static const char* szSuppMeth = ", PROPPATCH, PROPFIND, MKCOL, COPY, MOVE";
	
	static int iHeaderLen = strlen(szNewHeader);
	static int iSuppMethLen = strlen(szSuppMeth);
	static int iAnotherHeaderLen = strlen(szAnotherNewHeader);

	if(bOptionsRequest) {

		//Allocate data
		
		
		DWORD dwNewLen = pRawData->cbInData + iHeaderLen + iAnotherHeaderLen + iSuppMethLen * 2;
		
		char* szNy = new char[dwNewLen+1];
		strncpy(szNy, (char*)pRawData->pvInData, pRawData->cbInData);
		szNy[dwNewLen] = '\0';

		
		//Find the "Allow" header

		//**TODO: Check the URI and only add those methods which are in fact allowed


		char* szAllow = strstr(szNy, "\r\nAllow:");
		
		if(szAllow) {
			//Find the end of the header

			szAllow +=2;

			szAllow = strstr(szAllow, "\r\n");

			if(szAllow) {
				//Move everything up iSuppMethLen

				char* szEnd = szNy + pRawData->cbInData;

				while(szEnd >= szAllow) {
					*(szEnd + iSuppMethLen) = *szEnd;
					szEnd--;
				}

				//Insert our supported methods
				
				strncpy(szAllow, szSuppMeth, iSuppMethLen);
			}
		}

		//Find the "Public" header



		char* szPublic = strstr(szNy, "\r\nPublic:");
		
		if(szPublic) {
			//Find the end of the header

			szPublic +=2;

			szPublic = strstr(szPublic, "\r\n");

			if(szPublic) {
				//Move everything up iSuppMethLen

				char* szEnd = szNy + pRawData->cbInData + iSuppMethLen;

				while(szEnd >= szPublic) {
					*(szEnd + iSuppMethLen) = *szEnd;
					szEnd--;
				}

				//Insert our supported methods
				
				strncpy(szPublic, szSuppMeth, iSuppMethLen);
			}
		}



		//Insert new headers

		char* szBorder = strstr(szNy, "\r\n\r\n");

		if(szBorder) {
			szBorder += 2;
			strncpy(szBorder, szNewHeader, iHeaderLen);
			szBorder += iHeaderLen;

			strncpy(szBorder, szAnotherNewHeader, iAnotherHeaderLen);
			szBorder += iAnotherHeaderLen;

			strcpy(szBorder, "\r\n");
		}

		
		


		pCtxt->WriteClient(szNy, &dwNewLen);
		
		SAFE_FREE_ARRAY(szNy);

		bOptionsRequest = false;
		
		//return SF_STATUS_REQ_FINISHED_KEEP_CONN;
		return SF_STATUS_REQ_FINISHED;
	}


	return SF_STATUS_REQ_NEXT_NOTIFICATION;
}