//
// ACE.cpp
//
// Created by Marcelo Calbucci
//
// April/2000
//
// Compile with _WIN32_WINNT=0x0500 and _UNICODE defined

#include "ace.h"
#include "acl.h"
#include "sderror.h"
#include "windows.h"
#include "winbase.h"

// For comparison use, we don't care if the ACE was inherited or not
#define ALL_MEANINGFULL_ACE_FLAGS (~((DWORD)INHERITED_ACE))

// When setting the ACE Flags, some are valid and some are not
// depending of the ACE type...
#define ACE_FLAGS_FOR_DACL (OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | NO_PROPAGATE_INHERIT_ACE | INHERIT_ONLY_ACE)
#define ACE_FLAGS_FOR_SACL (ACE_FLAGS_FOR_DACL | SUCCESSFUL_ACCESS_ACE_FLAG | FAILED_ACCESS_ACE_FLAG)


// Global NULL GUID.
GUID gNullGuid = {0};

/////////////////
// Constructors
/////////////////

// Default constructor
CAce::CAce()
{
    m_eatAceType = AT_AccessAllowed;
    
    m_dwAceFlags = 0;
    m_dwMask = 0;
    
    m_guidObject = gNullGuid;
    m_guidInheritedObject = gNullGuid;

    m_pParentAcl = NULL;

    SetSDLastError(SD_SUCCESS);
}

// Copy constructor
CAce::CAce(const CAce & rhs)
{
    m_pParentAcl = NULL;
    SetAce(rhs);
}

// Constructor from raw ACE
CAce::CAce(const LPVOID pAce)
{
    m_pParentAcl = NULL;
    SetAce(pAce);
}

// Constructor with all parameters, and raw SID
CAce::CAce(eAceType eatType, const PSID pSid, DWORD dwMask, DWORD dwAceFlags, 
     const GUID & guidObjectType, const GUID & guidInheritedObjectType)
{
    m_pParentAcl = NULL;
    SetAce(eatType, pSid, dwMask, dwAceFlags, guidObjectType, guidInheritedObjectType);
}

// Constructor with all parameters, and CSid
CAce::CAce(eAceType eatType, const CSid & sid, DWORD dwMask, DWORD dwAceFlags, 
     const GUID & guidObjectType, const GUID & guidInheritedObjectType)
{
    m_pParentAcl = NULL;
    SetAce(eatType, sid, dwMask, dwAceFlags, guidObjectType, guidInheritedObjectType);
}

// Destructor
CAce::~CAce()
{
    // Nothing here
}

//////////////
// operators
//////////////

// Assignment operator
CAce & CAce::operator=(const CAce & rhs)
{
    SetAce(rhs);
    return *this;
}

// operator equal
BOOL CAce::operator==(const CAce & rhs) const
{
    if(&rhs==this)
        return TRUE;

    if(m_eatAceType != rhs.m_eatAceType)
        return FALSE;
    if(m_dwMask != rhs.m_dwMask)
        return FALSE;
    // two ACEs are identical if they have all flags identical, but we ignore the inheritance flag
    if( (m_dwAceFlags & ALL_MEANINGFULL_ACE_FLAGS) != (rhs.m_dwAceFlags & ALL_MEANINGFULL_ACE_FLAGS))
        return FALSE;
    if(m_guidObject != rhs.m_guidObject)
        return FALSE;
    if(m_guidInheritedObject != rhs.m_guidInheritedObject)
        return FALSE;

    if(m_Sid != rhs.m_Sid)
        return FALSE;

    return TRUE;
}

// operator not equal
BOOL CAce::operator!=(const CAce & rhs) const
{
    return !(operator==(rhs));
}

// Contains
// Check if the rhs is represented on 'this'. This is, if all members are identical
// and the mask of 'this' contains the mask of rhs.
// Parameter:
//   rhs: CAce to test
// Return:
//   True if rhs is represented in 'this'.
BOOL CAce::Contains(const CAce & rhs) const
{
    if(&rhs==this)
        return TRUE;

    if(m_eatAceType != rhs.m_eatAceType)
        return FALSE;
    if( (rhs.m_dwMask!=0) && ((m_dwMask & rhs.m_dwMask)!=rhs.m_dwMask) )
            return FALSE;
    if( (m_dwAceFlags & ALL_MEANINGFULL_ACE_FLAGS) != (rhs.m_dwAceFlags & ALL_MEANINGFULL_ACE_FLAGS))
        return FALSE;
    if(m_guidObject != rhs.m_guidObject)
        return FALSE;
    if(m_guidInheritedObject != rhs.m_guidInheritedObject)
        return FALSE;

    if(m_Sid != rhs.m_Sid)
        return FALSE;

    return TRUE;

}

////////
// Set
////////

// Set from CAce
void CAce::SetAce(const CAce & rhs)
{
    if(&rhs==this)
        return;

    Clear();

    m_eatAceType = rhs.m_eatAceType;
    
    m_dwAceFlags = rhs.m_dwAceFlags;
    m_dwMask = rhs.m_dwMask;
    
    m_Sid = rhs.m_Sid;
    
    m_guidObject = rhs.m_guidObject;
    m_guidInheritedObject = rhs.m_guidInheritedObject;
}

// Set from raw ACE
DWORD CAce::SetAce(const LPVOID pAce)
{
    Clear();

    if(pAce==NULL)
        return SetSDLastError(SD_INVALID_ACE);
    
    BOOL bObjectType = FALSE;
    
    BYTE AceType = *((BYTE *)pAce);
    switch(AceType) {
    case ACCESS_ALLOWED_OBJECT_ACE_TYPE:
        bObjectType = TRUE;
    case ACCESS_ALLOWED_ACE_TYPE:
        m_eatAceType = AT_AccessAllowed;
        break;
        
    case ACCESS_DENIED_OBJECT_ACE_TYPE:
        bObjectType = TRUE;
    case ACCESS_DENIED_ACE_TYPE:
        m_eatAceType = AT_AccessDenied;
        break;
        
    case SYSTEM_AUDIT_OBJECT_ACE_TYPE:
        bObjectType = TRUE;
    case SYSTEM_AUDIT_ACE_TYPE:
        m_eatAceType = AT_AuditAccess;
        break;
        
    default:
        return SetSDLastError(SD_INVALID_ACE_TYPE);
    }

    if(bObjectType) {
        PACCESS_ALLOWED_OBJECT_ACE pAA = (PACCESS_ALLOWED_OBJECT_ACE)pAce;
        PSID pSid = NULL;
        m_dwMask = pAA->Mask;
        SetFlags(pAA->Header.AceFlags);        
        if(pAA->Flags & ACE_OBJECT_TYPE_PRESENT) {
            m_guidObject = pAA->ObjectType;
            if(pAA->Flags & ACE_INHERITED_OBJECT_TYPE_PRESENT) {
                m_guidInheritedObject = pAA->InheritedObjectType;
                pSid = (PSID)&(pAA->SidStart);
            } else {
                pSid = (PSID)&(pAA->InheritedObjectType);
            }
        } else {
            if(pAA->Flags & ACE_INHERITED_OBJECT_TYPE_PRESENT) {
                m_guidInheritedObject = pAA->InheritedObjectType;
                pSid = (PSID)&(pAA->InheritedObjectType);
            } else {
                pSid = (PSID)&(pAA->ObjectType);
            }
        }
        return m_Sid.SetSid(pSid);
    } else {
        ACCESS_ALLOWED_ACE * pAA = (ACCESS_ALLOWED_ACE *)pAce;
        m_dwMask = pAA->Mask;
        SetFlags(pAA->Header.AceFlags);
        return m_Sid.SetSid(&(pAA->SidStart));
    }

}

// Set the Type
void CAce::SetType(eAceType eatType)
{
    // If the type is AuditAccess we limit for the SACL flags, otherwise the DACL flags
    if(eatType==AT_AuditAccess)
        SetFlags(GetFlags() & ACE_FLAGS_FOR_SACL);
    else
        SetFlags(GetFlags() & ACE_FLAGS_FOR_DACL);
    m_eatAceType = eatType;
    SetModified();
}

// Set the SID from a raw SID
DWORD CAce::SetSid(const PSID pSid)
{
    SetModified();
    return m_Sid.SetSid(pSid);
}

// Set the SID from a CSid
void CAce::SetSid(const CSid & sid)
{
    SetModified();
    m_Sid = sid;
}

// Set the Ace Flags
void CAce::SetFlags(DWORD dwAceFlags)
{
    SetModified();
    if(m_eatAceType==AT_AuditAccess)
        m_dwAceFlags = (dwAceFlags & ACE_FLAGS_FOR_SACL);
    else
        m_dwAceFlags = (dwAceFlags & ACE_FLAGS_FOR_DACL);
}

// Set the Mask
void CAce::SetMask(DWORD dwMask)
{
    SetModified();
    m_dwMask = dwMask;
}

// Set the Object Type
void CAce::SetObjectType(const GUID & guidObjectType)
{
    SetModified();
    m_guidObject = guidObjectType;
}

// Set the Inherited Object Type
void CAce::SetInheritedObjectType(const GUID & guidInheritedObjectType)
{
    SetModified();
    m_guidInheritedObject = guidInheritedObjectType;
}

// Set from all elements (using raw SID)
// Note, some values default to NULL:
//    dwAceFlags = 0
//    guidObjectType = gNullGuid
//    guidInheritedObjectType = gNullGuid
DWORD CAce::SetAce(eAceType eatType, const PSID pSid, DWORD dwMask, DWORD dwAceFlags, 
                   const GUID & guidObjectType, 
                   const GUID &guidInheritedObjectType)
{
    m_eatAceType = eatType;
    
    SetFlags(dwAceFlags);
    m_dwMask = dwMask;

    m_guidObject = guidObjectType;
    m_guidInheritedObject = guidInheritedObjectType;

    SetModified();
    
    return m_Sid.SetSid(pSid);
}

// Set from all elements (using CSid)
// Note, some values default to NULL:
//    dwAceFlags = 0
//    guidObjectType = gNullGuid
//    guidInheritedObjectType = gNullGuid
void CAce::SetAce(eAceType eatType, const CSid & sid, DWORD dwMask, DWORD dwAceFlags, 
                   const GUID & guidObjectType, 
                   const GUID &guidInheritedObjectType)
{
    m_eatAceType = eatType;
    
    SetFlags(dwAceFlags);
    m_dwMask = dwMask;
    
    m_guidObject = guidObjectType;
    m_guidInheritedObject = guidInheritedObjectType;

    m_Sid = sid;
    SetModified();
}

////////
// Get
////////

// GetRawAceType
// Return the raw ACE Type or 0xFF if unknown
BYTE CAce::GetRawAceType() const
{
    if(IsObjectTypeAce()) {
        switch(m_eatAceType) {
        case AT_AccessAllowed:
            return ACCESS_ALLOWED_OBJECT_ACE_TYPE;
        case AT_AccessDenied:
            return ACCESS_DENIED_OBJECT_ACE_TYPE;
        case AT_AuditAccess:
            return SYSTEM_AUDIT_OBJECT_ACE_TYPE;
        }
    } else {
        switch(m_eatAceType) {
        case AT_AccessAllowed:
            return ACCESS_ALLOWED_ACE_TYPE;
        case AT_AccessDenied:
            return ACCESS_DENIED_ACE_TYPE;
        case AT_AuditAccess:
            return SYSTEM_AUDIT_ACE_TYPE;
        }
    }
    return 0xFF;
}

// AddIntoAcl
// Add 'this' ACE into an ACL.
// Parameter:
//   pAcl: ACL that will receive this ACE. It must have enough space otherwise will fail
// Return:
//   Error code.
DWORD CAce::AddIntoAcl(PACL pAcl) const
{
    if(!pAcl)
        return SetSDLastError(SD_INVALID_ACL);

    // If the Mask is 0, no right is represented
    if(m_dwMask==0)         
        return SetSDLastError(SD_SUCCESS);

    // If there is not SID, ignore it
    if(m_Sid.GetSid()==NULL)
        return SetSDLastError(SD_SUCCESS); 
    
    const GUID * pObject;
    if(HasObjectType())
        pObject = &m_guidObject;
    else
        pObject = NULL;
    
    const GUID * pInherited;
    if(HasInheritedObjectType())
        pInherited = &m_guidInheritedObject;
    else
        pInherited = NULL;

    BOOL bObjectType = ((pObject!=NULL)||(pInherited!=NULL))?TRUE:FALSE;
    
    switch(m_eatAceType) {
    case AT_AccessAllowed:
        if(bObjectType) {
            if(!AddAccessAllowedObjectAce(pAcl, ACL_REVISION_DS, m_dwAceFlags, m_dwMask,
                                          (GUID *)pObject, (GUID *)pInherited, m_Sid.GetSid()))
                return SetSDLastError(SD_ACE_ADD_FAILURE_W32);
        } else {
            if(!AddAccessAllowedAceEx(pAcl, ACL_REVISION_DS, m_dwAceFlags, m_dwMask,
                                      m_Sid.GetSid()))
                return SetSDLastError(SD_ACE_ADD_FAILURE_W32);
        }
        break;
    case AT_AccessDenied:
        if(bObjectType) {
            if(!AddAccessDeniedObjectAce(pAcl, ACL_REVISION_DS, m_dwAceFlags, m_dwMask,
                                         (GUID *)pObject, (GUID *)pInherited, m_Sid.GetSid()))
                return SetSDLastError(SD_ACE_ADD_FAILURE_W32);
        } else {
            if(!AddAccessDeniedAceEx(pAcl, ACL_REVISION_DS, m_dwAceFlags, m_dwMask,
                m_Sid.GetSid()))
                return SetSDLastError(SD_ACE_ADD_FAILURE_W32);
        }
        break;
    case AT_AuditAccess:
        if(bObjectType) {
            if(!AddAuditAccessObjectAce(pAcl, ACL_REVISION_DS, m_dwAceFlags, m_dwMask,
                                        (GUID *)pObject, (GUID *)pInherited, m_Sid.GetSid(),
                                        FALSE, FALSE))
                return SetSDLastError(SD_ACE_ADD_FAILURE_W32);
        } else {
            if(!AddAuditAccessAceEx(pAcl, ACL_REVISION_DS, m_dwAceFlags, m_dwMask,
                                  m_Sid.GetSid(), FALSE, FALSE))
                return SetSDLastError(SD_ACE_ADD_FAILURE_W32);
        }
        break;
    default:
        return SetSDLastError(SD_INVALID_ACE_TYPE);
    }
    return SetSDLastError(SD_SUCCESS);
    
}

// LengthRequired
// Return the size (in bytes) required to store this ACE
DWORD CAce::LengthRequired() const
{
    // This values are present on all ACEs
    DWORD dwSize = sizeof(ACE_HEADER) + sizeof(ACCESS_MASK);

    // If the ACE has only objectType, need to add the size of one GUID.
    // If the ACE has only inheritedObjectType, need to add the size of one GUID.
    // If the ACE has an objectType and inheritedObjecType, need two GUIDs.
    // If objectType or inheritedObjectType is present, need to add the size of Flags (DWORD).
    if(m_guidObject!=gNullGuid) {
        if(m_guidInheritedObject!=gNullGuid) {
            dwSize += (sizeof(GUID) * 2) + sizeof(DWORD);
        } else {
            dwSize += sizeof(GUID) + sizeof(DWORD);
        }
    } else if(m_guidInheritedObject!=gNullGuid) {
        dwSize += sizeof(GUID) + sizeof(DWORD);
    }
    // Add the size of the SID.
    dwSize += m_Sid.LengthRequired();
    return dwSize;
}

// Clear
void CAce::Clear()
{
    SetSDLastError(SD_SUCCESS);

    m_dwAceFlags = 0;
    m_dwMask = 0;
    m_eatAceType = AT_AccessAllowed;
    m_guidInheritedObject = gNullGuid;
    m_guidObject = gNullGuid;

    m_Sid.Clear();
    SetModified();
}

// SetModified (protected)
void CAce::SetModified()
{
    // If this CAce is inside an CAcl, then we propagate modifications
    if(m_pParentAcl)
        m_pParentAcl->SetModified();
}

// SetParent (protected)
// This function should be called by the CAcl that owns this CAce.
void CAce::SetParent(CAcl * pAcl)
{
    m_pParentAcl = pAcl;
}