//
// SD.cpp
//
// Created by Marcelo Calbucci
//
// April/2000
//
// Compile with _WIN32_WINNT=0x0500 and _UNICODE defined

#include "sd.h"
#include "sderror.h"
#include "sddl.h"

// When comparing two SD, some bits are not important on the Control
#define SD_MEANINFULL_CONTROL_FOR_COMPARISON (SE_DACL_AUTO_INHERITED | SE_DACL_PROTECTED | \
                                                SE_SACL_AUTO_INHERITED | SE_SACL_PROTECTED)

//////////////////
// Constructors
//////////////////

// Default constructor
CSD::CSD()
{
    m_Control = 0;

    m_fModified = TRUE;
    m_pSDAbsolute = NULL;
    m_pSDRelative = NULL;

    m_dwAbsExtraDaclSize = 0;
    m_dwAbsExtraSaclSize = 0;

    m_dwRelExtraDaclSize = 0;
    m_dwRelExtraSaclSize = 0;
    m_fUseLocalFree = FALSE;

    m_Dacl.SetParentSD(this);
    m_Sacl.SetParentSD(this);
}

// Copy constructor
CSD::CSD(const CSD & rhs)
{
    m_fModified = TRUE;

    m_pSDAbsolute = NULL;
    m_pSDRelative = NULL;

    m_Dacl.SetParentSD(this);
    m_Sacl.SetParentSD(this);

    SetSD(rhs);
}

// Constructor from a raw SD
CSD::CSD(const PSECURITY_DESCRIPTOR pSd)
{
    m_fModified = TRUE;

    m_pSDAbsolute = NULL;
    m_pSDRelative = NULL;

    m_Dacl.SetParentSD(this);
    m_Sacl.SetParentSD(this);

    SetSD(pSd);
}

// Constructor from all parts of a SD in raw format
CSD::CSD(const PSID pOwner, const PSID pGroup,
        const PACL pDacl, const PACL pSacl)
{
    m_fModified = TRUE;

    m_pSDAbsolute = NULL;
    m_pSDRelative = NULL;

    m_Dacl.SetParentSD(this);
    m_Sacl.SetParentSD(this);

    m_Control = 0;

    SetSD(pOwner, pGroup, pDacl, pSacl);
}

// Constructor from all parts of a SD
CSD::CSD(const CSid & sidOwner, const CSid & sidGroup,
        const CAcl & Dacl, const CAcl & Sacl)
{
    m_fModified = TRUE;

    m_pSDAbsolute = NULL;
    m_pSDRelative = NULL;

    m_Dacl.SetParentSD(this);
    m_Sacl.SetParentSD(this);

    m_Control = 0;

    SetSD(sidOwner, sidGroup, Dacl, Sacl);
}

// Destructor
CSD::~CSD()
{
    delete m_pSDAbsolute;
    if(m_fUseLocalFree)
        LocalFree(m_pSDRelative); 
    else
        delete m_pSDRelative;
}

//////////////
// Operators
//////////////

// operator assignment
CSD & CSD::operator=(const CSD & rhs)
{
    SetSD(rhs);
    return *this;
}

// operator equal
BOOL CSD::operator==(const CSD & rhs) const
{
    if(&rhs==this)
        return TRUE;

    if((SD_MEANINFULL_CONTROL_FOR_COMPARISON & m_Control) !=
        (SD_MEANINFULL_CONTROL_FOR_COMPARISON & rhs.m_Control))
        return FALSE;

    if(m_SidOwner!=rhs.m_SidOwner)
	return FALSE;
    
    if(m_SidGroup!=rhs.m_SidGroup)
	return FALSE;
    
    if(m_Sacl != rhs.m_Sacl)
	return FALSE;
    
    if(m_Dacl != rhs.m_Dacl)
	return FALSE;
    
    return TRUE;
}

// operator not equal
BOOL CSD::operator!=(const CSD & rhs) const
{
    return !(operator==(rhs));
}

/////////
// Set
/////////

// Set from a raw SD
DWORD CSD::SetSD(const PSECURITY_DESCRIPTOR pSD)
{
    if(pSD) {
        if(m_pSDAbsolute==pSD || m_pSDRelative==pSD) // Self assignment
            return SetSDLastError(SD_SUCCESS);
    } else {
        // Empty SD is invalid
        Clear(); 
	return SetSDLastError(SD_INVALID_SD);
    } 
    
    Clear();
    if(!IsValidSecurityDescriptor(pSD)) {
	return SetSDLastError(SD_INVALID_SD);
    }
    
    DWORD dwRevision;
    // Get the Control
    if(!GetSecurityDescriptorControl(pSD, &m_Control, &dwRevision)) {
        return SetSDLastError(SD_GET_CONTROL_W32);
    }
    m_Control = m_Control & ALL_CONTROL_BITS; // Clear some useless bits

    DWORD dwError;
    BOOL bDefaulted;
    PSID pSid;

    // Get the Owner
    if(!GetSecurityDescriptorOwner(pSD, &pSid, &bDefaulted)) {
        Clear();
	return SetSDLastError(SD_GET_OWNER_WIN32);
    }
    dwError = m_SidOwner.SetSid(pSid);
    if(dwError!=SD_SUCCESS)
	return SetSDLastError(dwError);
    
    // Get the Group
    if(!GetSecurityDescriptorGroup(pSD, &pSid, &bDefaulted)) {
	Clear();
	return SetSDLastError(SD_GET_GROUP_WIN32);
    }
    dwError = m_SidGroup.SetSid(pSid);
    if(dwError!=SD_SUCCESS) {
	Clear();
	return SetSDLastError(dwError);
    }
    
    PACL pAcl;
    BOOL bPresent;
    
    // Get the DACL
    if(!GetSecurityDescriptorDacl(pSD, &bPresent, &pAcl, &bDefaulted)) {
	Clear();
	return SetSDLastError(SD_GET_DACL_WIN32);
    }
    if(bPresent) {
	dwError = m_Dacl.SetAcl(pAcl);
	if(dwError != SD_SUCCESS) {
	    Clear();
	    return SetSDLastError(dwError);
	}
    } else {
	m_Dacl.RemoveAllAces(TRUE);
    }
    
    // Get the SACL
    if(!GetSecurityDescriptorSacl(pSD, &bPresent, &pAcl, &bDefaulted)) {
	Clear();
	return SetSDLastError(SD_GET_SACL_WIN32);
    }
    if(bPresent) {
	dwError = m_Sacl.SetAcl(pAcl);
	if(dwError != SD_SUCCESS) {
	    Clear();
	    return SetSDLastError(dwError);
	}
    } else {
	m_Sacl.RemoveAllAces(TRUE);
    }
    
    return SetSDLastError(SD_SUCCESS);
}

// Set from a String SD
DWORD CSD::SetSDFromString(const wstring & wsStringSD)
{
    Clear();
    if(wsStringSD.size()==0) {
        return SetSDLastError(SD_INVALID_SD);
    }

    DWORD dwSize;
    if(!ConvertStringSecurityDescriptorToSecurityDescriptor(wsStringSD.c_str(),
                                SDDL_REVISION_1, &m_pSDRelative, &dwSize)) {
        return SetSDLastError(SD_CONVERT_STRING_SD_W32);
    }
    m_fUseLocalFree = TRUE;

    return SetSDFromRelative();
}

// Set from a CSD
void CSD::SetSD(const CSD & rhs)
{
    if(&rhs==this)
        return;

    Clear();

    m_Control = rhs.m_Control;

    m_Dacl = rhs.m_Dacl;
    m_Sacl = rhs.m_Sacl;
    m_SidOwner = rhs.m_SidOwner;
    m_SidGroup = rhs.m_SidGroup;
}

// Set from each part on raw format
DWORD CSD::SetSD(const PSID pOwner, const PSID pGroup,
		 const PACL pDacl, const PACL pSacl)
{

    DWORD dwError;
    if(pOwner) { // Owner
	dwError = m_SidOwner.SetSid(pOwner);
	if(dwError != SD_SUCCESS) {
	    Clear();
	    return SetSDLastError(dwError);
	}
    } else {
        m_SidOwner.Clear();
    }
    
    if(pGroup) { // Group
	dwError = m_SidGroup.SetSid(pGroup);
	if(dwError != SD_SUCCESS) {
	    Clear();
	    return SetSDLastError(dwError);
	}
    } else {
        m_SidGroup.Clear();
    }
    
    if(pDacl) { // DACL
	dwError = m_Dacl.SetAcl(pDacl);
	if(dwError != SD_SUCCESS) {
	    Clear();
	    return SetSDLastError(dwError);
	}
    } else {
	m_Dacl.RemoveAllAces(TRUE);
    }
    
    if(pSacl) { // SACL
	dwError = m_Sacl.SetAcl(pSacl);
	if(dwError != SD_SUCCESS) {
	    Clear();
	    return SetSDLastError(dwError);
	}
    } else {
	m_Sacl.RemoveAllAces(TRUE);
    }

    SetModified();
    
    return SetSDLastError(SD_SUCCESS);
}

// Set from each part
void CSD::SetSD(const CSid & sidOwner, const CSid & sidGroup,
		const CAcl & Dacl, const CAcl & Sacl)
{
    m_SidOwner = sidOwner;
    m_SidGroup = sidGroup;
    
    m_Dacl = Dacl;
    m_Sacl = Sacl;

    SetModified();
}

// Set the control bits
void CSD::SetControl(SECURITY_DESCRIPTOR_CONTROL Control)
{
    // Erase all bits that can be set
    m_Control = (m_Control & ~((WORD)ALL_CONTROL_BITS_OF_INTEREST));
    // Set the bits of interest with the parameter
    m_Control = m_Control | (Control & ALL_CONTROL_BITS_OF_INTEREST);
}

// Set the DACL
void CSD::SetDacl(const CAcl & Dacl)
{
    m_Dacl = Dacl;
}

// Set the DACL from raw ACL
DWORD CSD::SetDacl(const PACL pDacl)
{
    if(!pDacl) {
	m_Dacl.RemoveAllAces(TRUE);
        return SetSDLastError(SD_SUCCESS);
    } else {
        return m_Dacl.SetAcl(pDacl);
    }
}

// Set the SACL
void CSD::SetSacl(const CAcl & Sacl)
{
    m_Sacl = Sacl;
}

// Set the SACL from raw ACL
DWORD CSD::SetSacl(const PACL pSacl)
{
    if(!pSacl) {
	m_Sacl.RemoveAllAces(TRUE);
        return SetSDLastError(SD_SUCCESS);
    } else {
        return m_Sacl.SetAcl(pSacl);
    }
}

// Set Owner
void CSD::SetOwner(const CSid & sid)
{
    m_SidOwner = sid;
    SetModified();
}

// Set Owner from raw SID
DWORD CSD::SetOwner(const PSID pSid)
{
    SetModified();
    return m_SidOwner.SetSid(pSid);
}

// Set Owner and Group
void CSD::SetOwners(const CSid & sidOwner, const CSid & sidGroup)
{
    m_SidOwner = sidOwner;
    m_SidGroup = sidGroup;
    SetModified();
}

// Set Owner and Group from SIDs
DWORD CSD::SetOwners(const PSID pSidOwner, const PSID pSidGroup)
{
    m_SidGroup.Clear();

    DWORD dwError;
    dwError = m_SidOwner.SetSid(pSidOwner);
    if(dwError != SD_SUCCESS)
        return dwError;

    SetModified();
    return m_SidGroup.SetSid(pSidGroup);
}

// Set Group
void CSD::SetGroup(const CSid & sid)
{
    m_SidGroup.SetSid(sid);
    SetModified();
}

// Set Group from SID
DWORD CSD::SetGroup(const PSID pSid)
{	
    SetModified();
    return m_SidGroup.SetSid(pSid);
}

/////////
// Add
/////////

// AddAce
//   Add a CAce into the DACL or SACL
// Parameters:
//   Ace: CAce to be inserted
//   iPos: Position to add, less than zero insert at the end. (default to -1)
// Return: Error code.
DWORD CSD::AddAce(const CAce & Ace, int iPos)
{
    switch(Ace.GetType()) {
    case AT_AccessAllowed:
    case AT_AccessDenied:
        // If Access Allowed or Denied add on the DACL
        m_Dacl.AddAce(Ace, iPos);
        break;
    case AT_AuditAccess:
        // If Audit Access add on the SACL
        m_Sacl.AddAce(Ace, iPos);
        break;
    default:
        return SetSDLastError(SD_INVALID_ACE_TYPE);
    }
    CheckControlFlags();

    return SetSDLastError(SD_SUCCESS);
}

// AddAce
//   Add a CAce from a raw ACE
// Parameters:
//   pAce: Pointer to the ACE structure
//   iPos: Position to add, less than zero insert at the end. (default to -1)
// Return: Error code.
DWORD CSD::AddAce(const LPVOID pAce, int iPos)
{
    CAce Ace(pAce);
    if(GetSDLastError()!=SD_SUCCESS)
        return GetSDLastError();
    return AddAce(Ace, iPos);
}

// AddAccessAllowed
//   Add an Access Allowed CAce into the DACL.
// Parameters:
//   pSid: Pointer to the SID structure
//   dwMask: ACE Mask
//   dwAceFlags: ACE Flags (default to 0)
//   guidObjectType: The object-type GUID (default to NULL Guid)
//   guidInheritedObjectType: the inherited-object-type GUID (default to a NULL GUID)
//   iPos: Position to add it (default to -1, at the end)
// Return: Error code
DWORD CSD::AddAccessAllowed(const PSID pSid, DWORD dwMask, DWORD dwAceFlags, 
			    const GUID & guidObjectType, 
			    const GUID &guidInheritedObjectType,
			    int iPos)
{
    DWORD dwError;
    dwError = m_Dacl.AddAccessAllowed(pSid, dwMask, dwAceFlags, guidObjectType,
                                      guidInheritedObjectType, iPos);
    return dwError;
}

// AddAccessDenied
//   Add an Access Denied CAce into the DACL.
// Parameters:
//   pSid: Pointer to the SID structure
//   dwMask: ACE Mask
//   dwAceFlags: ACE Flags (default to 0)
//   guidObjectType: The object-type GUID (default to NULL Guid)
//   guidInheritedObjectType: the inherited-object-type GUID (default to a NULL GUID)
//   iPos: Position to add it (default to -1, at the end)
// Return: Error code
DWORD CSD::AddAccessDenied(const PSID pSid, DWORD dwMask, DWORD dwAceFlags, 
			   const GUID & guidObjectType, 
			   const GUID &guidInheritedObjectType,
			   int iPos)
{
    DWORD dwError;
    dwError = m_Dacl.AddAccessDenied(pSid, dwMask, dwAceFlags, guidObjectType,
                                     guidInheritedObjectType, iPos);
    return dwError;
}

// AddAuditAccess
//   Add an Audit Access ACE with the specified paramater into the SACL
// Parameters:
//   pSid: Pointer to the SID
//   dwMask: ACE Mask
//   dwAceFlags: ACE Flags (default to FAILED_ACCESS_ACE_FLAG)
//   guidObjectType: The object-type GUID (default to NULL Guid)
//   guidInheritedObjectType: the inherited-object-type GUID (default to a NULL GUID)
//   iPos: Position to add it (default to -1, at the end)
// Return: Error code
DWORD CSD::AddAuditAccess(const PSID pSid, DWORD dwMask, DWORD dwAceFlags, 
			  const GUID & guidObjectType, 
			  const GUID &guidInheritedObjectType,
			  int iPos)
{
    DWORD dwError;
    dwError = m_Sacl.AddAuditAccess(pSid, dwMask, dwAceFlags, guidObjectType,
                                    guidInheritedObjectType, iPos);
    return dwError;
}


// AddAccessAllowed
//   Add an Access Allowed ACE with the specified paramater into the DACL
// Parameters:
//   sid: CSid containing the SID
//   dwMask: ACE Mask
//   dwAceFlags: ACE Flags (default to 0)
//   guidObjectType: The object-type GUID (default to NULL Guid)
//   guidInheritedObjectType: the inherited-object-type GUID (default to a NULL GUID)
//   iPos: Position to add it (default to -1, at the end)
void CSD::AddAccessAllowed(const CSid & sid, DWORD dwMask, DWORD dwAceFlags, 
			   const GUID & guidObjectType, 
			   const GUID &guidInheritedObjectType,
			   int iPos)
{
    m_Dacl.AddAccessAllowed(sid, dwMask, dwAceFlags, guidObjectType,
                            guidInheritedObjectType, iPos);
}

// AddAccessDenied
//   Add an Access Denied ACE with the specified paramater into the DACL
// Parameters:
//   sid: CSid containing the SID
//   dwMask: ACE Mask
//   dwAceFlags: ACE Flags (default to 0)
//   guidObjectType: The object-type GUID (default to NULL Guid)
//   guidInheritedObjectType: the inherited-object-type GUID (default to a NULL GUID)
//   iPos: Position to add it (default to -1, at the end)
void CSD::AddAccessDenied(const CSid & sid, DWORD dwMask, DWORD dwAceFlags, 
			  const GUID & guidObjectType, 
			  const GUID &guidInheritedObjectType,
			  int iPos)
{
    m_Dacl.AddAccessDenied(sid, dwMask, dwAceFlags, guidObjectType,
                           guidInheritedObjectType, iPos);
}

// AddAuditAccess
//   Add an Audit Access ACE with the specified paramater into the SACL
// Parameters:
//   sid: CSid containing the SID
//   dwMask: ACE Mask
//   dwAceFlags: ACE Flags (default to FAILED_ACCESS_ACE_FLAG)
//   guidObjectType: The object-type GUID (default to NULL Guid)
//   guidInheritedObjectType: the inherited-object-type GUID (default to a NULL GUID)
//   iPos: Position to add it (default to -1, at the end)
// Return: Error code
void CSD::AddAuditAccess(const CSid & sid, DWORD dwMask, DWORD dwAceFlags, 
			 const GUID & guidObjectType, 
			 const GUID &guidInheritedObjectType,
			 int iPos)
{
    m_Sacl.AddAuditAccess(sid, dwMask, dwAceFlags, guidObjectType,
                          guidInheritedObjectType, iPos);
}

////////
// Get
////////

// Dacl
//   Return a reference to the CAcl that stores the DACL
CAcl & CSD::Dacl()
{
    return m_Dacl;
}

// Sacl
//   Return a reference to the CAcl that stores the SACL
CAcl & CSD::Sacl()
{
    return m_Sacl;
}

// GetOwner
//   Return the Owner
const CSid & CSD::GetOwner() const
{
    return m_SidOwner;
}

// GetGroup
//   Return the group
const CSid & CSD::GetGroup() const
{
    return m_SidGroup;
}

// GetSD
//   Return the raw SD (it can be on Absolute or Self-Relative)
const PSECURITY_DESCRIPTOR CSD::GetSD()
{
    if(m_pSDAbsolute) {
        return m_pSDAbsolute;
    } else if(m_pSDRelative) {
        return m_pSDRelative;
    }
    return GetSDAbsolute(0, 0);
}

// GetSDAbsolute
//   Return the the raw SD in Absolute format
const PSECURITY_DESCRIPTOR CSD::GetSDAbsolute()
{
    if(!m_pSDAbsolute) {
        return GetSDAbsolute(0,0);
    }
    return m_pSDAbsolute;

}

// GetSDAbsolute (protected)
//   Return the raw SD in Absolute format, and add some extra bytes at the end of DACL and SACL
const PSECURITY_DESCRIPTOR CSD::GetSDAbsolute(int iExtraDaclBytes, int iExtraSaclBytes)
{
    if(m_pSDAbsolute) {
        if(iExtraDaclBytes <= m_dwAbsExtraDaclSize && iExtraSaclBytes <= m_dwAbsExtraSaclSize) {
            // If we already have the SD, return it.
            return m_pSDAbsolute;
        }
        delete m_pSDAbsolute;
        m_pSDAbsolute = NULL;
    }

    PACL pAcl = NULL;
    PSID pSid = NULL;

    try {
        // Allocate space for SD structure
        m_pSDAbsolute = new SECURITY_DESCRIPTOR;
        if(m_pSDAbsolute==NULL) {
            SetSDLastError(SD_OUT_OF_MEMORY);
            return NULL;
        }

        // Initialize the SD
        if(!InitializeSecurityDescriptor(m_pSDAbsolute, SECURITY_DESCRIPTOR_REVISION)) {
            SetSDLastError(SD_INITIALIZE_SD_FAIL_W32);
            goto Fail;
        }

        // Set the Control
        SECURITY_DESCRIPTOR_CONTROL ControlBitsOfInterest = ALL_CONTROL_BITS_OF_INTEREST;
        SECURITY_DESCRIPTOR_CONTROL Control = m_Control & ControlBitsOfInterest;
        if(!SetSecurityDescriptorControl(m_pSDAbsolute, ControlBitsOfInterest, Control)) {
            SetSDLastError(SD_SET_CONTROL_W32);
            goto Fail;
        }

        // Set the Owner
        pSid = m_SidOwner.GetSid();
        if(pSid!=NULL) {
            if(!SetSecurityDescriptorOwner(m_pSDAbsolute, pSid, FALSE)) {
                SetSDLastError(SD_SET_OWNER_W32);
                goto Fail;
            }
        }

        // Set the Group
        pSid = m_SidGroup.GetSid();
        if(pSid!=NULL) {
            if(!SetSecurityDescriptorGroup(m_pSDAbsolute, pSid, FALSE)) {
                SetSDLastError(SD_SET_GROUP_W32);
                goto Fail;
            }
        }

        // Set the DACL
        pAcl = m_Dacl.GetAcl(iExtraDaclBytes);
        if(pAcl) {
            if(!SetSecurityDescriptorDacl(m_pSDAbsolute, TRUE, pAcl, FALSE)) {
                SetSDLastError(SD_SET_DACL_W32);
                goto Fail;
            }
        }

        // Set the SACL
        pAcl = m_Sacl.GetAcl(iExtraSaclBytes);
        if(pAcl) {
            if(!SetSecurityDescriptorSacl(m_pSDAbsolute, TRUE, pAcl, FALSE)) {
                SetSDLastError(SD_SET_SACL_W32);
                goto Fail;
            }
        }

    } catch(...) {
        delete m_pSDAbsolute;
        m_pSDAbsolute = NULL;
        throw;
    }

    SetSDLastError(SD_SUCCESS);

    m_dwAbsExtraDaclSize = iExtraDaclBytes;
    m_dwAbsExtraSaclSize = iExtraSaclBytes;

    m_fModified = FALSE;

    return m_pSDAbsolute;

Fail:
    delete m_pSDAbsolute;
    m_pSDAbsolute = NULL;

    return NULL;
}

// GetSDRelative
//   Return the raw SD in Self-Relative format
const PSECURITY_DESCRIPTOR CSD::GetSDRelative()
{
    if(!m_pSDRelative) {
        return GetSDRelative(0, 0);
    }
    return m_pSDRelative;

}

// GetSDRelative (protected)
//   Return the raw SD in Self-Relative format, and add some extra bytes 
//   at the end of DACL and SACL
const PSECURITY_DESCRIPTOR CSD::GetSDRelative(int iExtraDaclBytes, int iExtraSaclBytes)
{
    if(m_pSDRelative) {
        if(iExtraDaclBytes <= m_dwRelExtraDaclSize && iExtraSaclBytes <= m_dwRelExtraSaclSize) {
            // If we have the structure ready, return it.
            return m_pSDRelative;
        }
        if(m_fUseLocalFree)
            LocalFree(m_pSDRelative);
        else
            delete m_pSDRelative;
        m_pSDRelative = NULL;
    }

    // To create a Self-Relative SD, we need an absolute SD first...
    GetSDAbsolute(iExtraDaclBytes, iExtraSaclBytes);
    if(GetSDLastError()!=SD_SUCCESS)
        return NULL;

    DWORD dwSpace = LengthRequired() + iExtraDaclBytes + iExtraSaclBytes;
    try {
        // Allocate the space
        m_pSDRelative = new UCHAR[dwSpace];
        if(m_pSDRelative==NULL) {
            SetSDLastError(SD_OUT_OF_MEMORY);
            return NULL;
        }
        m_fUseLocalFree = FALSE;

        // Transform the Absolute into Self-Relative format
        if(!MakeSelfRelativeSD(m_pSDAbsolute, m_pSDRelative, &dwSpace)) {
            SetSDLastError(SD_MAKE_SELF_SD_W32);
            goto Fail;
        }
    } catch(...) {
        delete m_pSDRelative;
        m_pSDRelative = NULL;
        throw;
    }

    SetSDLastError(SD_SUCCESS);

    m_dwRelExtraDaclSize = iExtraDaclBytes;
    m_dwRelExtraSaclSize = iExtraSaclBytes;

    m_fModified = FALSE;

    return m_pSDRelative;

Fail:
    delete m_pSDRelative;
    m_pSDRelative = NULL;

    return NULL;
}

// GetSDAbsoluteCopy
//   Return a copy of the SD in Absolute format, that should be released with DeleteSD()
// Parameters:
//   iExtraDaclBytes: Amount of extra bytes to be added to the end of the DACL
//   iExtraSaclBytes: Amount of extra bytes to be added to the end of the SACL
// Return: Pointer to a SD structure.
PSECURITY_DESCRIPTOR CSD::GetSDAbsoluteCopy(int iExtraDaclBytes, int iExtraSaclBytes)
{	
    PSECURITY_DESCRIPTOR pSD = NULL;

    PACL pAcl = NULL;
    PSID pSid = NULL;

    try {
        // Allocate the structure
        pSD = new SECURITY_DESCRIPTOR;
        if(pSD==NULL) {
            SetSDLastError(SD_OUT_OF_MEMORY);
            return NULL;
        }

        // Initialize the SD
        if(!InitializeSecurityDescriptor(pSD, SECURITY_DESCRIPTOR_REVISION)) {
            SetSDLastError(SD_INITIALIZE_SD_FAIL_W32);
            delete pSD;
            return NULL;
        }

        // Set the Control
        SECURITY_DESCRIPTOR_CONTROL ControlBitsOfInterest = ALL_CONTROL_BITS_OF_INTEREST;
        SECURITY_DESCRIPTOR_CONTROL Control = m_Control & ControlBitsOfInterest;
        if(!SetSecurityDescriptorControl(m_pSDAbsolute, ControlBitsOfInterest, Control)) {
            SetSDLastError(SD_SET_CONTROL_W32);
            goto Fail;
        }

        // Set the Owner
        pSid = m_SidOwner.GetSidCopy();
        if(GetSDLastError()!=SD_SUCCESS) {
            goto Fail;
        }
        if(pSid!=NULL) {
            if(!SetSecurityDescriptorOwner(pSD, pSid, FALSE)) {
                delete pSid;
                SetSDLastError(SD_SET_OWNER_W32);
                goto Fail;
            }
            pSid = NULL; // Transfer the ownership
        }

        // Set the Group
        pSid = m_SidGroup.GetSidCopy();
        if(GetSDLastError()!=SD_SUCCESS) {
            goto Fail;
        }
        if(pSid!=NULL) {
            if(!SetSecurityDescriptorGroup(pSD, pSid, FALSE)) {
                delete pSid;
                SetSDLastError(SD_SET_GROUP_W32);
                goto Fail;
            }
            pSid = NULL; // Transfer the ownership
        }

        // Set the DACL
        pAcl = m_Dacl.GetAclCopy(iExtraDaclBytes);
        if(GetSDLastError()!=SD_SUCCESS) {
            goto Fail;
        }
        if(GetSDLastError()!=SD_SUCCESS) {
            delete pAcl;
            goto Fail;
        }
        if(pAcl) {
            if(!SetSecurityDescriptorDacl(pSD, TRUE, pAcl, FALSE)) {
                delete pAcl;
                SetSDLastError(SD_SET_DACL_W32);
                goto Fail;
            }
            pAcl = NULL; // Transfer the ownership
        }

        // Set the SACL
        pAcl = m_Sacl.GetAcl(iExtraSaclBytes);
        if(GetSDLastError()!=SD_SUCCESS) {
            goto Fail;
        }
        if(GetSDLastError()!=SD_SUCCESS) {
            delete pAcl;
            goto Fail;
        }
        if(pAcl) {
            if(!SetSecurityDescriptorSacl(m_pSDAbsolute, TRUE, pAcl, FALSE)) {
                delete pAcl;
                SetSDLastError(SD_SET_SACL_W32);
                goto Fail;
            }
            pAcl = NULL; // Transfer the ownership
        }

    } catch(...) {
        DeleteSD(pSD);
        delete pAcl;
        delete pSid;
        throw;
    }

    SetSDLastError(SD_SUCCESS);

    return pSD;

Fail:
    DeleteSD(pSD); 
    return NULL;
}

// GetSDRelativeCopy
//   Return a copy of the SD in Self-Relative format, that should be released with DeleteSD()
// Parameters:
//   iExtraDaclBytes: Amount of extra bytes to be added to the end of the DACL
//   iExtraSaclBytes: Amount of extra bytes to be added to the end of the SACL
// Return: Pointer to a SD structure.
PSECURITY_DESCRIPTOR CSD::GetSDRelativeCopy(int iExtraDaclBytes, int iExtraSaclBytes)
{
    // Generate the Absolute format...
    GetSDAbsolute(iExtraDaclBytes, iExtraSaclBytes);
    if(GetSDLastError()!=SD_SUCCESS)
        return NULL;

    PSECURITY_DESCRIPTOR pSD = NULL;

    DWORD dwSpace = LengthRequired() + iExtraDaclBytes + iExtraSaclBytes;
    try {
        // Allocate single block of memory
        pSD = new UCHAR[dwSpace];
        if(pSD==NULL) {
            SetSDLastError(SD_OUT_OF_MEMORY);
            return NULL;
        }

        // Convert from Absolute to Self-Relative.
        if(!MakeSelfRelativeSD(m_pSDAbsolute, pSD, &dwSpace)) {
            SetSDLastError(SD_MAKE_SELF_SD_W32);
            goto Fail;
        }
    } catch(...) {
        delete pSD;
        throw;
    }

    return pSD;

Fail:
    delete pSD;

    return NULL;
}

// GetStringSD
//   Return a SD in a string format.
// Parameters:
//   wsSD (out): Receive the string SD
//   si: The Security Information to be generated (default to SD_SI_ALL)
// Return: Error code.
DWORD CSD::GetStringSD(wstring &wsSD, SECURITY_INFORMATION si)
{
    wsSD.erase();

    // Get the SD 
    PSECURITY_DESCRIPTOR pSD = GetSD();
    if(!pSD)
        return GetSDLastError();

    LPTSTR szSD = NULL;
    try {
        ULONG ulSize;
        // Convert the raw SD into a string SD
        if(!ConvertSecurityDescriptorToStringSecurityDescriptor(pSD, SDDL_REVISION_1, 
                                                            si, &szSD, &ulSize)) {
            return SetSDLastError(SD_CONVERT_STRING_SD_W32);
        }
        wsSD = szSD;
    } catch(...) {
        LocalFree(szSD);
        throw;
    }
    LocalFree(szSD);
    return SetSDLastError(SD_SUCCESS);
    
}

// DeleteSD (static)
//   Free the content of a pointer to SD returned by GetSDAbsoluteCopy() or GetSDRelativeCopy().
void CSD::DeleteSD(PSECURITY_DESCRIPTOR pSD)
{
    if(!pSD)
        return;

    if(((SECURITY_DESCRIPTOR *)pSD)->Control & SE_SELF_RELATIVE) {
        // Self relative format is a single block of memory
        delete pSD;
    } else {
        // Absolute needs to free each component too.
        delete ((SECURITY_DESCRIPTOR *)pSD)->Owner;
        delete ((SECURITY_DESCRIPTOR *)pSD)->Group;
        delete ((SECURITY_DESCRIPTOR *)pSD)->Dacl;
        delete ((SECURITY_DESCRIPTOR *)pSD)->Sacl;
        delete pSD;
    }
    return;
}

/////////////
// Specials
/////////////

// RemoveInheritedAces
//   Remove the Inherited Aces from the SACL and DACL
void CSD::RemoveInheritedAces()
{
    m_Sacl.RemoveInheritedAces();
    m_Dacl.RemoveInheritedAces();
}

// CanonicalSort
//   Put the DACL on a Canonical format
void CSD::CanonicalSort()
{
    m_Dacl.CanonicalSort();
}

/////////
// Misc
/////////

// LengthRequired
//   Return the length required to hold the SD in Self-Relative format (in bytes).
DWORD CSD::LengthRequired() const
{
    DWORD dwSize = sizeof(SECURITY_DESCRIPTOR);
    dwSize += m_SidOwner.LengthRequired();
    dwSize += m_SidGroup.LengthRequired();
    dwSize += m_Dacl.LengthRequired();
    dwSize += m_Sacl.LengthRequired();

    return dwSize;
}

// Clear
void CSD::Clear()
{
    m_Control = 0;

    m_SidOwner.Clear();
    m_SidGroup.Clear();
    m_Sacl.RemoveAllAces(TRUE);
    m_Dacl.RemoveAllAces(TRUE);

    SetModified();
}

// GetFileSecurity
//   Get the SD from a file.
// Parameters:
//   szFileName: File to read the SD.
//   si: Security information to read (default to SI_SD_ALL)
// Return: Error code.
DWORD CSD::GetFileSecurity(LPCTSTR szFileName, SECURITY_INFORMATION si)
{
    Clear();

    DWORD dwMinLength = 0;
    // First call to get the size required
    ::GetFileSecurity(szFileName, si, m_pSDRelative, dwMinLength, &dwMinLength);

    if(dwMinLength==0)
        return SetSDLastError(SD_GET_FILE_SD_W32);

    // Allocate memory
    m_pSDRelative = new UCHAR[dwMinLength];
    if(m_pSDRelative==NULL)
        return SetSDLastError(SD_OUT_OF_MEMORY);
    m_fUseLocalFree = FALSE;

    // Fill the m_pSDRelative with the SD of the file
    if(!::GetFileSecurity(szFileName, si, m_pSDRelative, dwMinLength, &dwMinLength)) {
        delete m_pSDRelative;
        m_pSDRelative = NULL;
        return SetSDLastError(SD_GET_FILE_SD_W32);
    }

    // Populate the rest of the class with the information from the m_pSDRelative
    return SetSDFromRelative();    
}

// GetKernelObjectSecurity
//   Get the SD from a Kernel Object.
// Parameters:
//   hObject: Handle to the object.
//   si: Security information to read (default to SI_SD_ALL)
// Return: Error code.
DWORD CSD::GetKernelObjectSecurity(HANDLE hObject, SECURITY_INFORMATION si)
{
    Clear();

    DWORD dwMinLength = 0;

    // First call to get the size required
    ::GetKernelObjectSecurity(hObject, si, m_pSDRelative, 0, &dwMinLength);

    if(dwMinLength==0)
        return SetSDLastError(SD_GET_KERNEL_SD_W32);

    // Allocate the memory
    m_pSDRelative = new UCHAR[dwMinLength];
    if(m_pSDRelative==NULL)
        return SetSDLastError(SD_OUT_OF_MEMORY);
    m_fUseLocalFree = FALSE;

    // Fill the m_pSDRelative with the SD from the Kernel Object
    if(!::GetKernelObjectSecurity(hObject, si, m_pSDRelative, dwMinLength, &dwMinLength)) {
        delete m_pSDRelative;
        m_pSDRelative = NULL;
        return SetSDLastError(SD_GET_KERNEL_SD_W32);
    }

    return SetSDFromRelative();    
}

// GetUserObjectSecurity
//   Get the SD from a User Object.
// Parameters:
//   hObject: Handle to the object.
//   si: Security information to read (default to SI_SD_ALL)
// Return: Error code.
DWORD CSD::GetUserObjectSecurity(HANDLE hObject, SECURITY_INFORMATION si)
{
    Clear();

    DWORD dwMinLength = 0;

    // First call to get the size required
    ::GetUserObjectSecurity(hObject, &si, m_pSDRelative, 0, &dwMinLength);
    if(dwMinLength==0)
        return SetSDLastError(SD_GET_USER_SD_W32);

    // Allocate the memory
    m_pSDRelative = new UCHAR[dwMinLength];
    if(m_pSDRelative==NULL)
        return SetSDLastError(SD_OUT_OF_MEMORY);
    m_fUseLocalFree = FALSE;

    // Fill the m_pSDRelative with the SD from the User Object
    if(!::GetUserObjectSecurity(hObject, &si, m_pSDRelative, dwMinLength, &dwMinLength)) {
        delete m_pSDRelative;
        m_pSDRelative = NULL;
        return SetSDLastError(SD_GET_USER_SD_W32);
    }

    // Populate the rest of the class with the information from the m_pSDRelative
    return SetSDFromRelative();    
}

// GetServiceObjectSecurity
//   Get the SD from a Service object
// Parameters:
//   hService: Handle to the object.
//   si: Security information to read (default to SI_SD_ALL)
// Return: Error code.
DWORD CSD::GetServiceObjectSecurity(SC_HANDLE hService, SECURITY_INFORMATION si)
{
    Clear();

    DWORD dwMinLength = 0;

    // First call to get the size required
    ::QueryServiceObjectSecurity(hService, si, m_pSDRelative, 0, &dwMinLength);
    if(dwMinLength==0)
        return SetSDLastError(SD_GET_SERVICE_SD_W32);

    // Allocate the memory
    m_pSDRelative = new UCHAR[dwMinLength];
    if(m_pSDRelative==NULL)
        return SetSDLastError(SD_OUT_OF_MEMORY);
    m_fUseLocalFree = FALSE;

    // Fill the m_pSDRelative with the SD from the Service Object
    if(!::QueryServiceObjectSecurity(hService, si, m_pSDRelative, dwMinLength, &dwMinLength)) {
        delete m_pSDRelative;
        m_pSDRelative = NULL;
        return SetSDLastError(SD_GET_SERVICE_SD_W32);
    }

    // Populate the rest of the class with the information from the m_pSDRelative
    return SetSDFromRelative();    
}

// GetRegKeySecurity
//   Get the SD from a Registry Key
// Parameters:
//   hKey: Handle to key.
//   si: Security information to read (default to SI_SD_ALL)
// Return: Error code.
DWORD CSD::GetRegKeySecurity(HKEY hKey, SECURITY_INFORMATION si)
{
    Clear();

    DWORD dwMinLength = 0;

    // First call to get the size required
    RegGetKeySecurity(hKey, si, m_pSDRelative, &dwMinLength);
    if(dwMinLength==0)
        return SetSDLastError(SD_GET_REGKEY_SD_W32);

    // Allocate the memory
    m_pSDRelative = new UCHAR[dwMinLength];
    if(m_pSDRelative==NULL)
        return SetSDLastError(SD_OUT_OF_MEMORY);
    m_fUseLocalFree = FALSE;

    // Fill the m_pSDRelative with the SD from the Registry Key
    if(RegGetKeySecurity(hKey, si, m_pSDRelative, &dwMinLength)!=ERROR_SUCCESS) {
        delete m_pSDRelative;
        m_pSDRelative = NULL;
        return SetSDLastError(SD_GET_REGKEY_SD_W32);
    }

    // Populate the rest of the class with the information from the m_pSDRelative
    return SetSDFromRelative();    
}

// SetFileSecurity
//   Set the SD on the file.
// Parameters:
//   szFileName: File to set the SD
//   si: Security information to update (default to SI_SD_ALL)
// Return: Error code.
DWORD CSD::SetFileSecurity(LPCTSTR szFileName, SECURITY_INFORMATION si)
{
    if(!szFileName || szFileName[0] ==0)
        return SetSDLastError(SD_INVALID_FILE);

    if(si==0)
        return SetSDLastError(SD_NO_SECURITY_INFORMATION_SET);

    // Get the SD
    PSECURITY_DESCRIPTOR pSD = GetSD();
    if(pSD==NULL) {
        return GetSDLastError();
    }

    // Assign the SD to the file
    if(!::SetFileSecurity(szFileName, si, pSD)) {
        return SetSDLastError(SD_SET_FILE_SECURITY_W32);
    }
    return SetSDLastError(SD_SUCCESS);

}

// SetKernelObjectSecurity
//   Set the SD on the Kernel Object.
// Parameters:
//   hObject: Handle to the object to set the SD
//   si: Security information to update (default to SI_SD_ALL)
// Return: Error code.
DWORD CSD::SetKernelObjectSecurity(HANDLE hObject, SECURITY_INFORMATION si)
{
    if(hObject==INVALID_HANDLE_VALUE)
        return SetSDLastError(SD_INVALID_HANDLE);

    if(si==0)
        return SetSDLastError(SD_NO_SECURITY_INFORMATION_SET);

    // Get the SD
    PSECURITY_DESCRIPTOR pSD = GetSD();
    if(pSD==NULL) {
        return GetSDLastError();
    }

    // Assign the SD to the object
    if(!::SetKernelObjectSecurity(hObject, si, pSD)) {
        return SetSDLastError(SD_SET_KERNEL_OBJECT_SECURITY_W32);
    }
    return SetSDLastError(SD_SUCCESS);

}

// SetUserObjectSecurity
//   Set the SD on the User Object.
// Parameters:
//   hObject: Handle to the object to set the SD
//   si: Security information to update (default to SI_SD_ALL)
// Return: Error code.
DWORD CSD::SetUserObjectSecurity(HANDLE hObject, SECURITY_INFORMATION si)
{
    if(hObject==INVALID_HANDLE_VALUE)
        return SetSDLastError(SD_INVALID_HANDLE);

    if(si==0)
        return SetSDLastError(SD_NO_SECURITY_INFORMATION_SET);

    // Get the SD
    PSECURITY_DESCRIPTOR pSD = GetSD();
    if(pSD==NULL) {
        return GetSDLastError();
    }

    // Assign the SD to the object
    if(!::SetUserObjectSecurity(hObject, &si, pSD)) {
        return SetSDLastError(SD_SET_USER_OBJECT_SECURITY_W32);
    }
    return SetSDLastError(SD_SUCCESS);

}

// SetServiceObjectSecurity
//   Set the SD on the Serivce Object.
// Parameters:
//   hService: Handle to the object to set the SD
//   si: Security information to update (default to SI_SD_ALL)
// Return: Error code.
DWORD CSD::SetServiceObjectSecurity(SC_HANDLE hService, SECURITY_INFORMATION si)
{
    if(hService==INVALID_HANDLE_VALUE)
        return SetSDLastError(SD_INVALID_HANDLE);

    if(si==0)
        return SetSDLastError(SD_NO_SECURITY_INFORMATION_SET);

    // Get the SD
    PSECURITY_DESCRIPTOR pSD = GetSD();
    if(pSD==NULL) {
        return GetSDLastError();
    }

    // Assign the SD to the object
    if(!::SetServiceObjectSecurity(hService, si, pSD)) {
        return SetSDLastError(SD_SET_SERVICE_OBJECT_SECURITY_W32);
    }
    return SetSDLastError(SD_SUCCESS);


}

// SetRegKeySecurity
//   Set the SD on a Registry Key.
// Parameters:
//   hKey: Handle to the RegKey to set the SD
//   si: Security information to update (default to SI_SD_ALL)
// Return: Error code.
DWORD CSD::SetRegKeySecurity(HKEY hKey, SECURITY_INFORMATION si)
{
    if(si==0)
        return SetSDLastError(SD_NO_SECURITY_INFORMATION_SET);

    // Get the SD
    PSECURITY_DESCRIPTOR pSD = GetSD();
    if(pSD==NULL) {
        return GetSDLastError();
    }

    // Assign the SD to the RegKey
    if(!::RegSetKeySecurity(hKey, si, pSD)) {
        return SetSDLastError(SD_REG_SET_KEY_SECURITY_W32);
    }
    return SetSDLastError(SD_SUCCESS);
}


// SetModified (protected)
//   Erase the m_pSDAbsolute and m_pSDRelative if the SD change.
void CSD::SetModified()
{
    if(m_fModified)
        return;

    m_fModified = TRUE;

    if(m_pSDAbsolute) {
        delete m_pSDAbsolute;
        m_pSDAbsolute = NULL;
    }

    if(m_pSDRelative) {
        if(m_fUseLocalFree) {
            LocalFree(m_pSDRelative);
            m_fUseLocalFree = FALSE;
        } else {
           delete m_pSDRelative;
        }
        m_pSDRelative = NULL;
    }

    m_dwAbsExtraDaclSize = m_dwAbsExtraSaclSize = 0;
    m_dwRelExtraDaclSize = m_dwRelExtraSaclSize = 0;
}

// SetSDFromRelative (protected)
//   Populate all members from this instance with the content of m_pSDRelative
DWORD CSD::SetSDFromRelative()
{
    if(m_pSDRelative==NULL) {
        return SetSDLastError(SD_SUCCESS);
    }

    // Get the Control
    DWORD dwRevision;
    if(!GetSecurityDescriptorControl(m_pSDRelative, &m_Control, &dwRevision)) {
        return SetSDLastError(SD_GET_CONTROL_W32);
    }
    m_Control = m_Control & ALL_CONTROL_BITS; // Clear some useless bits

    DWORD dwError;
    BOOL bDefaulted;
    PSID pSid;
    
    // Construct the Owner
    if(!GetSecurityDescriptorOwner(m_pSDRelative, &pSid, &bDefaulted)) {
        dwError = SD_GET_OWNER_WIN32;
        goto Fail;
    }
    dwError = m_SidOwner.SetSid(pSid);
    if(dwError!=SD_SUCCESS)
	goto Fail;
    
    // Construct the Group
    if(!GetSecurityDescriptorGroup(m_pSDRelative, &pSid, &bDefaulted)) {
	dwError = SD_GET_GROUP_WIN32;
        goto Fail;
    }
    dwError = m_SidGroup.SetSid(pSid);
    if(dwError!=SD_SUCCESS) {
        goto Fail;
    }
    
    PACL pAcl;
    BOOL bPresent;
    
    // Construct the DACL
    if(!GetSecurityDescriptorDacl(m_pSDRelative, &bPresent, &pAcl, &bDefaulted)) {
	dwError = SD_GET_DACL_WIN32;
        goto Fail;
    }
    if(bPresent) {
	dwError = m_Dacl.SetAcl(pAcl);
	if(dwError != SD_SUCCESS) {
            goto Fail;
	}
    } else {
	m_Dacl.RemoveAllAces(TRUE);
    }
    
    // Construct the SACL
    if(!GetSecurityDescriptorSacl(m_pSDRelative, &bPresent, &pAcl, &bDefaulted)) {
	dwError = SD_GET_SACL_WIN32;
        goto Fail;
    }
    if(bPresent) {
	dwError = m_Sacl.SetAcl(pAcl);
	if(dwError != SD_SUCCESS) {
            goto Fail;
	}
    } else {
	m_Sacl.RemoveAllAces(TRUE);
    }

    SetModified();

    return SetSDLastError(SD_SUCCESS);

Fail:
    Clear();
    return SetSDLastError(dwError);
}

// CheckControlFlags
//   Populate the m_Control with the right bits (for read-only purpose).
//   This function is called from CAcl.
void CSD::CheckControlFlags()
{
    // Set/Reset the Control bits
    if(m_Dacl.IsPresent())
        m_Control = m_Control | SE_DACL_PRESENT;
    else
        m_Control = m_Control & (~(SECURITY_DESCRIPTOR_CONTROL)SE_DACL_PRESENT);

    if(m_Sacl.IsPresent())
        m_Control = m_Control | SE_SACL_PRESENT;
    else
        m_Control = m_Control & (~(SECURITY_DESCRIPTOR_CONTROL)SE_SACL_PRESENT);
}
