//
// SID.cpp
//
// Created by Marcelo Calbucci
//
// April/2000
//
// Compile with _WIN32_WINNT=0x0500 and _UNICODE defined

#include "sderror.h"
#include "sid.h"
#include "windows.h"
#include "sddl.h"


#ifdef _DEBUG
#define GETACCOUNTNAME { CSid::GetAccountName(m_pSid, m_wsName); }
#else
#define GETACCOUNTNAME { }
#endif

// This is the per Thread variable that will hold the last error
__declspec(thread) DWORD gdwSDLastError = SD_SUCCESS;

////////////////
// Constructors
////////////////

// Default constructor
CSid::CSid()
{
    m_fUseLocalFree = FALSE;
    m_pSid = NULL;
    SetSDLastError(SD_SUCCESS);
}

// Copy constructor
CSid::CSid(const CSid & rhs)
{
    m_pSid = NULL;
    SetSid(rhs);
}

// Constructor for eWellKnownSids (see sid.h)
CSid::CSid(eWellKnownSids eSid)
{
    m_pSid = NULL;
    SetSid(eSid);
}

// Constructor from a raw SID
CSid::CSid(const PSID pSid)
{
    m_pSid = NULL;
    SetSid(pSid);
}

// Destructor
CSid::~CSid()
{
    if(m_fUseLocalFree) {
        LocalFree(m_pSid);
    } else {
        delete m_pSid;
    }
}

//////////////
// operators
//////////////

// Operator assignment
CSid & CSid::operator=(const CSid & rhs)
{
    SetSid(rhs);
    return *this;
}

// operator equal
BOOL CSid::operator==(const CSid & rhs) const
{
    if(&rhs==this)
        return TRUE;

    if(!m_pSid && !rhs.m_pSid)
        return TRUE; // We assume if they are both empty, they are equal

    if(m_pSid && rhs.m_pSid) {
        return EqualSid(m_pSid, rhs.m_pSid);
    }
    return FALSE;
}

// operator not equal
BOOL CSid::operator!=(const CSid & rhs) const
{
    return !(operator==(rhs));
}

///////
// Set
///////

// Set from CSid
void CSid::SetSid(const CSid & rhs)
{

    if(&rhs==this)
        return;

    Clear();

    SetSDLastError(SD_SUCCESS);

    m_pSid = rhs.GetSidCopy();
    GETACCOUNTNAME;
}

// Set from raw SID
DWORD CSid::SetSid(const PSID pSid)
{
    if(pSid==NULL) {
        Clear();
        return SetSDLastError(SD_INVALID_SID);
    }
    if(pSid==m_pSid)
        return SetSDLastError(SD_SUCCESS);

    Clear();
    
    if(!IsValidSid(pSid)) {
        return SetSDLastError(SD_INVALID_SID);
    }
    
    int iSidSize = GetLengthSid(pSid);
    m_pSid = (PSID)new UCHAR[iSidSize];
    if(!m_pSid)
        return SetSDLastError(SD_OUT_OF_MEMORY);

    if(!CopySid(iSidSize, m_pSid, pSid)) {
        delete m_pSid;
        m_pSid = NULL;
        return SetSDLastError(SD_COPYSID_W32);
    }
    GETACCOUNTNAME;
    return GetSDLastError();
}

// Set from String SID
DWORD CSid::SetSidFromString(const wstring & wsStringSid)
{
    Clear();
    
    if(wsStringSid.size()==0) {
        return SetSDLastError(SD_INVALID_SID);
    }

    if(!ConvertStringSidToSid(wsStringSid.c_str(), &m_pSid)) {
        m_pSid = NULL;
        return SetSDLastError(SD_INVALID_SID);
    }
    m_fUseLocalFree = TRUE;            
    
    GETACCOUNTNAME;
    return GetSDLastError();
}

// Set from Account Name
DWORD CSid::SetSidFromName(const wstring & wsAccountName, LPCTSTR lpszServer)
{
    Clear();
    
    if(wsAccountName.size()==0) {
        return SetSDLastError(SD_INVALID_SID);
    }

    m_pSid = GetAccountSid(wsAccountName, lpszServer);
    if(!m_pSid)
        return GetSDLastError();
    m_fUseLocalFree = FALSE;
    
    GETACCOUNTNAME;
    return GetSDLastError();
}

// Set from Well-Known SID
DWORD CSid::SetSid(eWellKnownSids eSid)
{
    // These are the authorities that we're going to need
    static SID_IDENTIFIER_AUTHORITY siaNTAuth = SECURITY_NT_AUTHORITY;
    static SID_IDENTIFIER_AUTHORITY siaWorld = SECURITY_WORLD_SID_AUTHORITY;
    static SID_IDENTIFIER_AUTHORITY siaCreator = SECURITY_CREATOR_SID_AUTHORITY;

    Clear();

    SID_IDENTIFIER_AUTHORITY sia;
    DWORD dwSubAuth[8]; // Sub-Authorities (up to 8)
    ZeroMemory(dwSubAuth, sizeof(DWORD) * 8);
    int iTotalSubAuth = 1;

    PSID pSid = NULL;

    try {
        // The maximum Sub-Authority count that we use here is 2
        pSid = new UCHAR[GetSidLengthRequired(2)];
        if(!pSid)
            return SetSDLastError(SD_OUT_OF_MEMORY);

        switch(eSid) {
        case WKS_Self:              // Self
            sia = siaNTAuth;
            dwSubAuth[0] = SECURITY_PRINCIPAL_SELF_RID;
            break;
        case WKS_Everyone:          // Everyone (or World)
            sia = siaWorld;
            dwSubAuth[0] = SECURITY_WORLD_RID;
            break;
        case WKS_AuthenticatedUsers:// Authenticated Users
            sia = siaNTAuth;
            dwSubAuth[0] = SECURITY_AUTHENTICATED_USER_RID;
            break;
        case WKS_CreatorOwner:      // Creator Owner
            sia = siaCreator;
            dwSubAuth[0] = SECURITY_CREATOR_OWNER_RID;
            break;
        case WKS_Administrators:    // Administrators
            sia = siaNTAuth;
            dwSubAuth[0] = SECURITY_BUILTIN_DOMAIN_RID;
            dwSubAuth[1] = DOMAIN_ALIAS_RID_ADMINS;
            iTotalSubAuth = 2;
            break;
        case WKS_AccountOperators:  // Account Operators
            sia = siaNTAuth;
            dwSubAuth[0] = SECURITY_BUILTIN_DOMAIN_RID;
            dwSubAuth[1] = DOMAIN_ALIAS_RID_ACCOUNT_OPS;
            iTotalSubAuth = 2;
            break;
        case WKS_BackupOperators:   // Backup Operators
            sia = siaNTAuth;
            dwSubAuth[0] = SECURITY_BUILTIN_DOMAIN_RID;
            dwSubAuth[1] = DOMAIN_ALIAS_RID_BACKUP_OPS;
            iTotalSubAuth = 2;
            break;
        case WKS_Guests:            // Guests
            sia = siaNTAuth;
            dwSubAuth[0] = SECURITY_BUILTIN_DOMAIN_RID;
            dwSubAuth[1] = DOMAIN_ALIAS_RID_GUESTS;
            iTotalSubAuth = 2;
            break;
        case WKS_PrintOperators:    // Print Operators
            sia = siaNTAuth;
            dwSubAuth[0] = SECURITY_BUILTIN_DOMAIN_RID;
            dwSubAuth[1] = DOMAIN_ALIAS_RID_PRINT_OPS;
            iTotalSubAuth = 2;
            break;
        case WKS_Users:             // Users
            sia = siaNTAuth;
            dwSubAuth[0] = SECURITY_BUILTIN_DOMAIN_RID;
            dwSubAuth[1] = DOMAIN_ALIAS_RID_USERS;
            iTotalSubAuth = 2;
            break;

        // Not really "Well-Known SID" but useful
        case WKS_CurrentProcessSid:     // Context of Current Process 
        case WKS_CurrentThreadSid:      // Context of Current Thread
            return GetTokenSid(eSid);
        default:
            return SetSDLastError(SD_INTERNAL_ERROR);
            break;
        }
        // Initialize the SID and set the Authority
        if(!InitializeSid(pSid, &sia, iTotalSubAuth)) {
            delete pSid;
            return SetSDLastError(SD_INITIALIZE_SID_W32);
        }
        // Set the Sub-Authorities
        for(int i = 0; i < iTotalSubAuth; i++)
            *(GetSidSubAuthority(pSid, i)) = dwSubAuth[i];

    } catch(...) {
        delete pSid;
        throw;        
    }
    m_pSid = pSid;

    GETACCOUNTNAME;
    return GetSDLastError();
}

////////
// Get
////////

// GetSidCopy
// Get a copy of the SID and transfer the ownership to the caller
// Parameters:
//   iMinSpace = Minimum space to be allocated for that SID (default to 0)
// Return:
//   A copy of the SID and it should be deallocated with DeleteSid(pSid);
PSID CSid::GetSidCopy(int iMinSpace) const
{
    if(m_pSid==NULL)
        return NULL;
    
    int iSidSize = GetLengthSid(m_pSid);
    if(iSidSize < iMinSpace)
        iSidSize = iMinSpace;

    PSID pSid = NULL;

    try {
        pSid = (PSID)new UCHAR[iSidSize];
        if(!pSid) {
            SetSDLastError(SD_OUT_OF_MEMORY);
            return NULL;
        }

        if(!CopySid(iSidSize, pSid, m_pSid)) {
            delete pSid;
            SetSDLastError(SD_COPYSID_W32);
            return NULL;
        }
    } catch(...) {
        delete pSid;
        throw;
    }
    SetSDLastError(SD_SUCCESS);

    return pSid;
}

// GetStringSid
// Get the SID in a String format
// Paramter:
//   wsSid (out): Receive the SID in a string format if successful
// Return
//   Error code
DWORD CSid::GetStringSid(wstring & wsSid) const
{
    wsSid.erase();
    if(m_pSid==NULL) {
        return SetSDLastError(SD_INVALID_SID);
    }

    wchar_t * szStringSid = NULL;

    if(!ConvertSidToStringSid(m_pSid, &szStringSid)) {
        return SetSDLastError(SD_CONVERT_SID_TO_STRING_SID_W32);
    }
    try {
        wsSid = szStringSid;
    } catch(...) {
        LocalFree(szStringSid);
        throw;
    }
    LocalFree(szStringSid);
    return SetSDLastError(SD_SUCCESS);
}


// DeleteSid (static)
// Delete a SID allocated with GetSidCopy()
// Parameter:
//   pSid: Pointer to the SID
void CSid::DeleteSid(PSID pSid)
{
    delete pSid;
}

// LengthRequired
// Return the size of the SID
DWORD CSid::LengthRequired() const
{
    return m_pSid!=NULL?GetLengthSid(m_pSid):0;
}

// IsSamePrefix
// Check if two SIDs have the same prefix.
// Parameter:
//   rhs: Second sid to test agains 'this'
// Return:
//   True if they have the same prefix
BOOL CSid::IsSamePrefix(const CSid & rhs) const
{
    if(&rhs==this)
        return TRUE;

    // If one of them are NULL, they are not the same Prefix
    if(!m_pSid || !rhs.m_pSid)
        return FALSE;

    // Count the number of Sub Authorities
    int iSubAuth1, iSubAuth2;
    iSubAuth1 = *GetSidSubAuthorityCount(m_pSid);
    iSubAuth2 = *GetSidSubAuthorityCount(rhs.m_pSid);

    // Equal Prefix only works if the number of SubAuthorities is identical
    if(iSubAuth1 == iSubAuth2) {
        return EqualPrefixSid(m_pSid, rhs.m_pSid);
    }

    BOOL bResult = FALSE;
    PSID pSid = NULL;
    if((iSubAuth1 + 1)==  iSubAuth2) {
        // If there is 1 extra SubAuthority on the rhs than on this, add it
        PSID pSid = NULL;
        try {
            pSid = GetSidCopy(LengthRequired() + 16);
            if(!pSid) {
                return FALSE;
            }
            ((SID *)pSid)->SubAuthorityCount++;
            ((SID *)pSid)->SubAuthority[iSubAuth1] = 0;
            bResult = EqualPrefixSid(pSid, rhs.m_pSid);
        } catch(...) {
            delete pSid;
            throw;
        }
    }
    
    if(iSubAuth1 ==  (iSubAuth2 + 1)) {
        // If there is 1 extra SubAuthority on this than on rhs, add it
        PSID pSid = NULL;
        try {
            pSid = rhs.GetSidCopy(rhs.LengthRequired() + 16);
            if(!pSid) {
                return FALSE;
            }
            ((SID *)pSid)->SubAuthorityCount++;
            ((SID *)pSid)->SubAuthority[iSubAuth2] = 0;
            bResult = EqualPrefixSid(m_pSid, pSid);
        } catch(...) {
            delete pSid;
            throw;
        }
    }

    delete pSid;
    return bResult;
}

// Clear
void CSid::Clear()
{
    if(m_pSid) {
        if(m_fUseLocalFree)
            LocalFree(m_pSid);
        else
            delete m_pSid;
        m_pSid = NULL;
    }
    m_fUseLocalFree = FALSE;
#if _DEBUG
    m_wsName.erase();
#endif
    SetSDLastError(SD_SUCCESS);
}


// GetAccountName (static)
// Parameters:
//   pSid: Pointer to the SID to get the account.
//   wsAccountName (out): Receive the account name on format 'Domain\Account'
//   lpszServerName: Server to look for the SID (default is NULL)
// Return
//   TRUE if we succeeded find the account
BOOL CSid::GetAccountName(const PSID pSid, wstring &wsAccountName, LPCTSTR lpszServer)
{
    wsAccountName.erase();

    if(!pSid) {
        SetSDLastError(SD_INVALID_SID);
        return FALSE;
    }

    DWORD dwNameSize = 256;
    DWORD dwDomainSize = 256;

    TCHAR szName[256];
    TCHAR szDomain[256];
    
    szName[0] = 0;
    szDomain[0] = 0;
    
    SID_NAME_USE snu;

    if(!LookupAccountSid(lpszServer, pSid, szName, &dwNameSize, szDomain, &dwDomainSize, &snu)) {
        SetSDLastError(SD_LOOKUP_ACCOUNT_SID_W32);
        return FALSE;
    }

    wsAccountName = szDomain;
    if(szDomain[0] && szName[0])
        wsAccountName += L"\\";
    wsAccountName += szName;
    
    SetSDLastError(SD_SUCCESS);
    
    return TRUE;
}

// GetAccountSid (static)
// Return the SID of an account
// Parameters:
//   wsAccountName: Account that you want to find the SID
//   lpszServer: Server name to look for the Account (default to NULL)
// Return:
//   If success, return pointer to a SID and it should be freed with DeleteSid(pSid)
//   otherwise, NULL.
PSID CSid::GetAccountSid(const wstring & wsAccountName, LPCTSTR lpszServer)
{
    PSID pSid = NULL;
    try {
        DWORD dwDomainSize = 256;
        TCHAR szDomain[256];

        DWORD dwSidSize = 64;
        pSid = (PSID)new UCHAR[dwSidSize];
        if(!pSid) {
            SetSDLastError(SD_OUT_OF_MEMORY);
            return NULL;
        }            

        SID_NAME_USE snu;
        
        if(!LookupAccountName(lpszServer, wsAccountName.c_str(), pSid, &dwSidSize, 
                              szDomain, &dwDomainSize, &snu)) {
            SetSDLastError(SD_LOOKUP_ACCOUNT_NAME_W32);
            delete pSid;
            return NULL;
        }
    } catch(...) {
        delete pSid;
        throw;
    }

    SetSDLastError(SD_SUCCESS);
    return pSid;    
}

// GetTokenSid (protected)
// Get the SID of the current context of the Thread or Process
DWORD CSid::GetTokenSid(eWellKnownSids eSid)
{
    HANDLE TokenHandle;

    switch(eSid) {
    case WKS_CurrentThreadSid:
        if(OpenThreadToken(GetCurrentThread(), TOKEN_READ, FALSE, &TokenHandle)) {
            // Quit the switch if we could read the token
            break;
        } else {
            if(::GetLastError()!=ERROR_NO_TOKEN)
                return SetSDLastError(SD_OPEN_THREAD_TOKEN_W32);
            // If the error is ERROR_NO_TOKEN, we read the Process token
        }
        // fall thru
    case WKS_CurrentProcessSid:
        if(!OpenProcessToken(GetCurrentProcess(), TOKEN_READ, &TokenHandle)) {
            return SetSDLastError(SD_OPEN_PROCESS_TOKEN_W32);
        }
        break;
    default:
        return SetSDLastError(SD_INTERNAL_ERROR);
    }
    
    UCHAR Buff[128];

    PTOKEN_USER ptu = (PTOKEN_USER)Buff;
    DWORD dwLength;
    
    if(!GetTokenInformation(TokenHandle, TokenUser, ptu, sizeof(Buff), &dwLength)) {
        return SetSDLastError(SD_GET_TOKEN_INFORMATION_W32);
    }

    SetSDLastError(SD_SUCCESS);
    return SetSid(ptu->User.Sid);
}
